import logging
import os
import sys
import time
import subprocess
import traceback
import zipfile
import ssl

from validateMoreChecks import CasaManager

logger = None

APUAT_CWD = os.getcwd()
APUAT_SUPPORT_DATA_DIR = os.path.join(APUAT_CWD, 'apuat-support-data')
APUAT_LOG_DIR = os.path.join(APUAT_SUPPORT_DATA_DIR, 'log')
APUAT_REPORT_DIR = os.path.join(APUAT_SUPPORT_DATA_DIR, 'report')
APUAT_LOG_FILE = os.path.join(APUAT_LOG_DIR, 'apuat.log')
APUAT_SUPPORT_CONFIG = os.path.join(APUAT_CWD, 'apuatSupportBundleConfig.xml')
VALIDATE_MORE_CHECKS_PATH = os.path.join(APUAT_CWD, 'validateMoreChecks.py')
REPORT_STATIC_FILES_PATH = os.path.join(APUAT_CWD, 'report')

logger = None


def setupLogger(logDir):
    global logger
    logger = logging.getLogger('apuat-b2b')
    logger.setLevel(logging.DEBUG)
    # set log dir
    if not os.path.isdir(logDir):
        try:
            os.makedirs(logDir)
        except EnvironmentError as e:
            logger.error('Failed to make log dir "{0}". Exception: {1}'.format(logDir, str(e)))
            sys.exit(1)
    # create the file logger
    filename = os.path.join(logDir, 'apuat_install_{0}.log'.format(time.strftime("%Y%m%d-%H%M%S")))
    fh = logging.handlers.RotatingFileHandler(filename, 'a', 1048576, 5)
    fh.setLevel(logging.INFO)
    fileFormatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(module)s - %(funcName)s - %(message)s')
    fh.setFormatter(fileFormatter)
    logger.addHandler(fh)
    # create the console logger
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    consoleFormatter = logging.Formatter('%(levelname)s - %(message)s')
    ch.setFormatter(consoleFormatter)
    logger.addHandler(ch)


def runScript(scriptCommand, acceptableReturnCodes=None, suppressErrors=False, universal_newlines=True,
              runningDirectory=None, waitForExit=True, env=None):
    """
    Runs a script
    :param scriptCommand:
    :param scriptWithArgs the script and arguments to run
    :param acceptableReturnCodes a list of return codes acceptable upon completion of running the script. Default is [0].
    :return tuple (boolean-script ran successfully, int-exit code, str-stderr, str-stdout)
    """
    logger.debug('Script command: "{0}"'.format(scriptCommand))
    processPipe = subprocess.Popen(scriptCommand, shell=True, cwd=runningDirectory, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE, env=env, universal_newlines=universal_newlines)
    if waitForExit:
        (processStdOut, processStdErr) = processPipe.communicate()
        logger.debug('stdout: "{0}"'.format(processStdOut))
        logger.debug('stderr: "{0}"'.format(processStdErr))
        logger.debug('exit code: {0}'.format(processPipe.returncode))
        if not acceptableReturnCodes:
            acceptableReturnCodes = [0]
        if processPipe.returncode not in acceptableReturnCodes:
            message = 'Script command: "{0}" failed with exit code: {1}'.format(scriptCommand,
                                                                                processPipe.returncode)
            if suppressErrors:
                logger.debug(message)
            else:
                logger.error(message)
            return False, processPipe.returncode, processStdErr, processStdOut
        return True, processPipe.returncode, processStdErr, processStdOut
    else:
        logger.debug('Do not wait for script: "{0}" to complete'.format(scriptCommand))


def handleException(message, exception):
    logger.error(message)
    logger.error(str(exception))
    traceback.print_exc()


def createNecessaryDirs():
    for path in [APUAT_SUPPORT_DATA_DIR, APUAT_LOG_DIR, APUAT_REPORT_DIR]:
        try:
            os.mkdir(path)
        except OSError as ex:
            handleException('(APUAT) failed to create directory {0}.'.format(path), ex)

    logger.debug('(APUAT) APUAT_SUPPORT_DATA_DIR dir: {0}'.format(APUAT_SUPPORT_DATA_DIR))
    logger.debug('(APUAT) APUAT_LOG_DIR dir: {0}'.format(APUAT_LOG_DIR))
    logger.debug('(APUAT) APUAT_REPORT_DIR dir: {0}'.format(APUAT_REPORT_DIR))


def runPreUpgradeChecks():
    logger.info('Starting pre-upgrade validation checks')
    logger.debug('VMWARE_PYTHON_BIN {0}'.format(os.environ['VMWARE_PYTHON_BIN']))
    logger.debug('VALIDATE_MORE_CHECKS_PATH {0}'.format(VALIDATE_MORE_CHECKS_PATH))
    logger.debug('APUAT_REPORT_DIR {0}'.format(APUAT_REPORT_DIR))
    commandToRun = '{0} {1} {2}'.format(os.environ['VMWARE_PYTHON_BIN'],
                                        VALIDATE_MORE_CHECKS_PATH,
                                        APUAT_REPORT_DIR)
    runScript(commandToRun)
    logger.info('Successfully completed pre-upgrade validation checks')


def runDisabledMetricsCheck(javaPath, vropsBase):
    apuatJarPath = os.path.join(APUAT_CWD, 'apuat-jar', 'apuat-jar-1.0-SNAPSHOT-jar-with-dependencies.jar')
    if not os.path.isfile(apuatJarPath):
        handleException('(APUAT) Failed to find jar: {0}'.format(apuatJarPath))

    apuatClassPath = os.path.join(APUAT_CWD, 'apuat-classes')
    try:
        os.mkdir(apuatClassPath)
    except OSError as ex:
        handleException('(APUAT) Failed to create directory: {0}.'.format(apuatClassPath), ex)

    logger.info('(APUAT) Extracting jar: {0}'.format(apuatJarPath))
    ## APUAT-TODO what exception this throws ?
    with zipfile.ZipFile(apuatJarPath) as zipper:
        zipper.extractall(apuatClassPath)

    vcopsClassPath = os.path.join(vropsBase, 'common', 'lib') + os.sep + '*:' + os.path.join(vropsBase, 'controller', 'plugins', 'vcops-view-server-1.0-SNAPSHOT.jar')
    classPath = '{0}:{1}'.format(apuatClassPath, vcopsClassPath)
    bcFipsDeps = ''
    if os.environ['IS_FIPS']:
        bcFipsDeps = {
            "$ALIVE_BASE/common/lib/bc-fips-current.jar",
            "$ALIVE_BASE/common/lib/bcpkix-fips-current.jar",
            "$ALIVE_BASE/common/lib/bctls-fips-current.jar"
        }
        bcFipsDeps = ' -Xbootclasspath/a:' + ':'.join(bcFipsDeps)
    cmd = '{0} -cp \"{1}\" -DAPUAT_LOGFILE_FULLPATH={2} -DAPUAT_REPORT_BASEDIR={3} {4} -Dfile.encoding=UTF-8 com.vmware.apuat.PreUpgradeAssessorMain'.format(
        javaPath, classPath, APUAT_LOG_FILE, APUAT_REPORT_DIR, bcFipsDeps)

    return runScript(cmd, runningDirectory=APUAT_CWD)


def getEnvironmentVariableAbsValue(envVar):
    """
    Gets the abs path of an env var
    :param envVar The environment variable to lookup, this must be a file path
    :return String abs path value
    """
    logger.debug('Looking up value for environment variable: {0}'.format(envVar))
    try:
        value = os.path.abspath(os.environ[envVar])
    except KeyError as ex:
        handleException(
            '(APUAT) Failed to get value of environment variable: {0}.'.format(envVar), ex)

    logger.debug('(APUAT) Environment variable: {0} value: {1}'.format(envVar, value))
    return value


def generateSupportBundleConfig():
    with open(APUAT_SUPPORT_CONFIG, 'w') as configFile:
        configFile.write('<?xml version="1.0" encoding="ISO-8859-1" ?>\n')
        configFile.write('<config>\n')
        configFile.write(' <MaxSupportBundles>5</MaxSupportBundles>\n')
        configFile.write(' <daysFilter>1</daysFilter>\n')
        configFile.write('  <entries>\n')
        configFile.write('   <entry>\n')
        configFile.write('    <destDir>apuat-data</destDir>\n')
        configFile.write('    <fileset type="logs" subtypes="all,others" sizes="full,light">\n')
        configFile.write('     <srcDir>{0}</srcDir>\n'.format(APUAT_SUPPORT_DATA_DIR))
        configFile.write('   </fileset>\n')
        configFile.write('   </entry>\n')
        configFile.write('  </entries>\n')
        configFile.write('</config>\n')


def generateSupportBundle():
    generate_support_bundle_script = os.path.join(APUAT_CWD, 'apuatSupportBundleGenerator.py')
    try:
        if not os.path.isfile(APUAT_SUPPORT_CONFIG):
            logger.error(
                'Cannot generate support bundle as the script: "{0}" does not exist'.format(APUAT_SUPPORT_CONFIG))
            return False

        logger.debug('Attempting to generate support bundle')
        command = '{0} {1} -a create --config={2} -m'.format((os.environ['VMWARE_PYTHON_BIN'] or ''),
                                                             generate_support_bundle_script,
                                                             os.path.join(APUAT_CWD, APUAT_SUPPORT_CONFIG))
        runScript(command)
        logger.debug('Completed generating support bundle')
        return True
    except Exception as e:
        handleException('Cannot generate support bundle. Command is "{0}"'.format(command), e)
        return False


def main():
    try:
        setupLogger('/var/log/vmware/vcops')
    except Exception as setup_logs_exception:
        print('Failed to setup logs: {0}'.format(setup_logs_exception))
        sys.exit(1)

    logger.info('(APUAT) Start processing.')
    logger.info('(APUAT) Starting pre-upgrade checks.')
    try:
        createNecessaryDirs()
        runPreUpgradeChecks()
        runScript('cp {0}/* {1} '.format(REPORT_STATIC_FILES_PATH, APUAT_REPORT_DIR))
        logger.info('(APUAT) Finished pre-upgrade checks successfully.')
    except Exception as e:
        handleException("(APUAT) Exception when running pre-upgrade checks.", e)

    logger.info('(APUAT) Starting disabled metrics check.')
    isAdmin = False
    try:
        isAdmin = CasaManager().ping_node_to_find_admin(logger=logger)
    except ssl.SSLError as e:
        handleException("(APUAT) Certificate is expired cannot perform disabled metric check.", e)
    except Exception as e:
        handleException("(APUAT) Exception when checking if the node is admin.", e)

    if isAdmin:
        logger.info("Node is admin, will run disabled metrics check.")
        javaHome = getEnvironmentVariableAbsValue('VMWARE_JAVA_HOME')
        javaBin = os.path.join(javaHome, 'bin', 'java')
        vropsBase = getEnvironmentVariableAbsValue('VCOPS_BASE')
        (result, returnCode, stderr, stdout) = runDisabledMetricsCheck(javaBin, vropsBase)
        if not result:
            logger.error('(APUAT) Disabled metric check has failed: {0} {1}'.format(returnCode, stderr))
        logger.info("(APUAT) Finished disabled metric check successfully.")
    else:
        logger.info("(APUAT) Skipping disabled metric check for the node.")

    generateSupportBundleConfig()
    if generateSupportBundle():
        logger.info("(APUAT) Finished process successfully.")
        #remove pak_service.json file
        location = '/usr/lib/vmware-casa-b2b/casa-b2b'
        file = 'pak_service.json'
        path = os.path.join(location, file)
        if os.path.exists(location + '/' + file):
            os.remove(path)
            logger.info("%s has been removed successfully" %file)
        else:
            logger.info(f"Can't remove {file} file. Reason: the file does not exist")
        sys.exit(0)
    else:
        logger.info("(APUAT) Could not generate support bundle.")
        sys.exit(1)


if __name__ == "__main__":
    main()
