#!/usr/bin/python -cprint 'ERROR: You must invoke this script using $VMWARE_PYTHON_BIN'; import sys; sys.exit(5)
# Copyright 2017 VMware, Inc.  All rights reserved. -- VMware Confidential
# Updates a state file

__author__ = "VMware, Inc."

import logging
import logging.handlers
import os
import re
import platform
import time
import json
from optparse import OptionParser
import sys
import datetime
import shutil
try:
    from collections import OrderedDict
except ImportError:
    OrderedDict = None


class EnvironmentInfo(object):
    log_file_name = None
    log_level = None
    state_file_name = None

    @classmethod
    def newInstance(cls):
        if cls.IsWindows():
            item = Windows()
        else:
            item = Linux()
        item.log_file_name = 'pak_state_updater_{0}.log'.format(time.strftime('%Y%m%d-%H%M%S'))
        item.state_file_name = 'state_file.json'
        item.log_level = logging.DEBUG
        return item

    @classmethod
    def IsWindows(cls):
        """Returns boolean value if this is running on Windows or not"""
        isPlatformWindows = re.search(r'windows', platform.system(), re.IGNORECASE)
        return isPlatformWindows


class Windows(EnvironmentInfo):
    def __init__(self):
        super(Windows, self).__init__()
        self.logDir = os.path.join(os.environ['STORAGE'], 'log', 'vcops')


class Linux(EnvironmentInfo):
    def __init__(self):
        super(Linux, self).__init__()
        self.logDir = '/var/log/vmware/vcops'


def setup_logger():
    # setup the logger
    global logger
    logger = logging.getLogger('b2b')
    logger.setLevel(logging.DEBUG)
    envInfo = EnvironmentInfo.newInstance()
    # set log dir
    if not os.path.isdir(envInfo.logDir):
        try:
            os.makedirs(envInfo.logDir)
        except EnvironmentError as e:
            print('Failed to make log dir "{0}". Exception: {1}'.format(envInfo.logDir, str(e)))
            sys.exit(1)
    # create the file logger
    filename = os.path.join(envInfo.logDir, 'vcops_upgrade_{0}.log'.format(time.strftime("%Y%m%d-%H%M%S")))
    global b2bStatusFile
    b2bStatusFile = os.path.join(envInfo.logDir, 'vcops_upgrade_{0}.json'.format(time.strftime("%Y%m%d-%H%M%S")))
    global b2bTimingStatusFile
    b2bTimingStatusFile = os.path.join(envInfo.logDir, 'vcops_upgrade_timing_{0}.json'.format(time.strftime("%Y%m%d-%H%M%S")))
    global b2bStateFile
    b2bStateFile = os.path.join(envInfo.logDir, 'vcops_upgrade_state.json')
    mode = 'a'
    maxBytes=1048576
    backupCount=5
    fh = logging.handlers.RotatingFileHandler(filename, mode, maxBytes, backupCount)
    fh.setLevel(logging.DEBUG)
    fileFormatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(module)s - %(funcName)s - %(message)s')
    fh.setFormatter(fileFormatter)
    logger.addHandler(fh)
    # create the console logger
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    consoleFormatter = logging.Formatter('%(levelname)s - %(message)s')
    ch.setFormatter(consoleFormatter)
    logger.addHandler(ch)


class Action(object):
    def __init__(self, action_name, parent_name=None, start_time=None, stop_time=None, readable_start_time=None,
                 readable_stop_time=None, result=None, result_description=None, friendly_name=None, log_files=None):
        self.action_name = action_name
        self.start_time = start_time
        self.stop_time = stop_time
        self.parent_name = parent_name
        self.readable_start_time = readable_start_time
        self.readable_stop_time = readable_stop_time
        self.result = result
        self.result_description = result_description
        self.friendly_name = friendly_name
        self.log_files = log_files
        if not self.log_files:
            self.log_files = []

    @staticmethod
    def return_times():
        current_time = time.time()
        readable_current_time = datetime.datetime.fromtimestamp(current_time).strftime('%Y-%m-%d %H:%M:%S')
        return int(current_time * 1000), readable_current_time

    def set_start_time(self):
        current_time, readable_current_time = Action.return_times()
        logger.debug('Set start time to: %s (%s) for action: %s', current_time, readable_current_time, self.action_name)
        self.start_time = current_time
        self.readable_start_time = readable_current_time

    def set_stop_time(self):
        current_time, readable_current_time = Action.return_times()
        logger.debug('Set STOP time to: %s (%s) for action: %s', current_time, readable_current_time, self.action_name)
        self.stop_time = current_time
        self.readable_stop_time = readable_current_time

    def set_parent_name(self, parent_name):
        logger.debug('Setting parent to: %s for action: %s', parent_name, self.action_name)
        self.parent_name = parent_name

    def set_result(self, result):
        logger.debug('Setting result to: %s for action: %s', result, self.action_name)
        self.result = result

    def set_result_description(self, result_description):
        logger.debug('Setting parent to: %s for action: %s', result_description, self.action_name)
        self.result_description = result_description

    def return_json(self):
        return self.__dict__


class DbInfo(object):
    def __init__(self, name, size):
        self.name = name
        self.size = size

    def return_json(self):
        return self.__dict__

class StateDoc(object):
    template_state_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'updateCoordinator.json')

    def __init__(self, state_file_path):
        self.state_file_path = None
        self.state_doc = None
        self.action_list = []
        self.state_file_path = state_file_path
        if not os.path.isfile(self.state_file_path):
            self.deploy_state_template()
        self.set_state_doc()
        self.load_actions_list()

    def deploy_state_template(self):
        logger.debug('Entering deploy the state file template: "%s" to: "%s"', StateDoc.template_state_file, self.state_file_path)
        if os.path.isfile(StateDoc.template_state_file):
            try:
                shutil.copyfile(StateDoc.template_state_file, self.state_file_path)
            except EnvironmentError:
                error_message = 'Failed to copy template state file: "{}" to: "{}"'.format(StateDoc.template_state_file, self.state_file_path)
                logger.exception(error_message)
                raise error_message

    def from_json_file(self):
        """
        :rtype B2bState
        """
        self.state_doc = None
        logger.debug('state file: "%s"', self.state_file_path)
        if os.path.isfile(self.state_file_path):
            with open(self.state_file_path) as state_file:
                if OrderedDict:
                    self.state_doc = json.load(state_file, object_pairs_hook=OrderedDict)
                else:
                    self.state_doc = json.load(state_file)

    def setup_initial_state_doc(self):
        """
        :rtype b2bState
        """
        if OrderedDict:
            self.state_doc = json.loads('{"actions": []}', object_pairs_hook=OrderedDict)
        else:
            self.state_doc = json.loads('{"actions": []}')
        return self.state_doc

    def set_state_doc(self):
        if not self.state_doc:
            self.from_json_file()
        if not self.state_doc:
            self.setup_initial_state_doc()
        return self.state_doc

    def load_actions_list(self):
        self.set_state_doc()
        for action in self.state_doc['actions']:
            self.action_list.append(Action(**action))

    def ensure_action_exists(self, action_to_update):
        if not any(action.action_name == action_to_update for action in self.action_list):
            logger.debug('Adding the action: %s to the action list', action_to_update)
            self.action_list.append(Action(action_to_update))

    def update_start_time(self, action_to_update):
        self.ensure_action_exists(action_to_update)
        for action in self.action_list:
            if action.action_name == action_to_update:
                action.set_start_time()

    def update_stop_time(self, action_to_update):
        self.ensure_action_exists(action_to_update)
        for action in self.action_list:
            if action.action_name == action_to_update:
                action.set_stop_time()

    def update_parent(self, action_to_update, parent_name):
        self.ensure_action_exists(action_to_update)
        for action in self.action_list:
            if action.action_name == action_to_update:
                action.set_parent_name(parent_name)

    def update_result(self, action_to_update, result):
        self.ensure_action_exists(action_to_update)
        for action in self.action_list:
            if action.action_name == action_to_update:
                action.set_result(result)

    def update_result_description(self, action_to_update, result_description):
        self.ensure_action_exists(action_to_update)
        for action in self.action_list:
            if action.action_name == action_to_update:
                action.set_result_description(result_description)

    def update_db_phase_sizes(self, db_phase, sizes):
        logger.debug('Entering update DB phase sizes, phase: %s', db_phase)
        self.set_state_doc()
        if db_phase not in self.state_doc['db_sizes']:
            raise KeyError('DB Sizes key: "{0}" missing and cannot be updated'.format(db_phase))

        # load the DBs list
        dbs = []
        for db in self.state_doc['db_sizes'][db_phase]:
            dbs.append(DbInfo(**db))

        # cycle through the list of DBs and update the sizes.  If a size is passed in sizes that isn't in the template, it will be ignored
        for db in dbs:
            if db.name in sizes:
                logger.debug('Found DB: "%s", update the size to: %d', db.name, sizes[db.name])
                db.size = sizes[db.name]
                del sizes[db.name]
            else:
                logger.debug('DB: "%s" was not found in the list of sizes to update, skip updating value', db.name)

        # write sizes skipped (hopefully there are none)
        for db, size in sizes.iteritems():
            logger.warning('Cannot write DB size: "%s" to the phase: "%s" with the size: %d, because it does not exist',
                           db, db_phase, size)

        # update the state document
        self.state_doc['db_sizes'][db_phase] = []
        for db in dbs:
            self.state_doc['db_sizes'][db_phase].append(db.return_json())

        self.write_state_file()
        logger.debug('Completed updating DB phase sizes')

    def return_state_json(self):
        self.state_doc['actions'] = []
        for action in self.action_list:
            json = action.return_json()
            self.state_doc['actions'].append(json)

    def write_state_doc_actions_update(self):
        self.return_state_json()
        self.write_state_file()

    def write_state_file(self):
        try:
            with open(self.state_file_path, 'w') as working_state_file:
                json.dump(self.state_doc, working_state_file, indent=4, sort_keys=True)
        except EnvironmentError:
            logger.exception('Failed to write to the state file: "%s"', self.state_file_path)

    def return_current_action(self):
        """
        Looking at the start and stop time for each action, and returns the action for whatever was updated last
        :rtype Action
        """
        logger.debug('Entering return current action')
        current_time = None
        current_action = None
        for action in self.action_list:
            logger.debug('Checking action: %s', action.action_name)
            if action.start_time or action.stop_time:
                action_max_time = max(action.start_time, action.stop_time)
                if action_max_time > current_time:
                    current_action = action
                    current_time = action_max_time
                    logger.debug('Setting action to check: %s with the current time: %d', current_action.action_name, current_time)
                else:
                    logger.debug('Action: %s with the time: %d is older then the current action time: %d', action.action_name, action_max_time, current_time)

            else:
                logger.debug('Action: %s does not have a start or stop time, skip it', action.action_name)
        logger.info('Current action: %s has the start time: %d and end time: %d', current_action.action_name, current_action.start_time, current_action.stop_time)
        return current_action

def parse_args():
    usage = "usage: %prog"
    parser = OptionParser(usage)
    parser.add_option ("-a", "--action_name",
                       dest="action_name",
                       default="",
                       help="action to update")
    parser.add_option ("-f", "--state_file_path",
                       dest="state_file_path",
                       default="",
                       help="The path to the state file to update")
    parser.add_option ("-p", "--action_parent",
                       dest="action_parent",
                       default="",
                       help="parent action")
    parser.add_option ("-r", "--set_start_time",
                       action="store_true",
                       dest="set_start_time",
                       default=False,
                       help="Determines if the start time should be set")
    parser.add_option ("-e", "--result",
                       dest="result",
                       default="",
                       help="result of the action")
    parser.add_option ("-u", "--result_description",
                       dest="result_description",
                       default="",
                       help="result description of the action")
    parser.add_option ("-o", "--set_stop_time",
                       action="store_true",
                       dest="set_stop_time",
                       default=False,
                       help="Determines if the stop time should be set")

    (options, args) = parser.parse_args()

    logger.debug('input params:')
    logger.debug('  - action name: "%s"', options.action_name)
    logger.debug('  - state file path: "%s"', options.state_file_path)
    logger.debug('  - action parent: "%s"', options.action_parent)
    logger.debug('  - set start time: %s', options.set_start_time)
    logger.debug('  - set stop time: %s', options.set_stop_time)
    logger.debug('  - result: %s', options.result)
    logger.debug('  - result description: %s', options.result_description)
    return options


def main():
    raise NotImplementedError('Not currently implemented')
    setup_logger()
    options = parse_args()
    try:
        logger.debug('Entering update action info for action: "%s" with the state file: "%s" and the action parent: "%s"',
                     options.action_name, options.state_file_path, options.action_parent)
        update_state_file(options.state_file_path, options.action_name, options.action_parent, options.set_start_time,
                          options.set_stop_time, options.result, options.result_description)
    except EnvironmentError:
        logger.exception('Update action failed, exiting')
        sys.exit(1)


def set_logger(logger_in):
    global logger
    logger = logger_in


def verify_logger():
    if not logger:
        raise NameError('logger is not set')


def verify_state_file_exists(state_file_path):
    if not state_file_path or state_file_path == "":
        error_message = 'State file was not passed'
        logger.exception(error_message)
        raise EnvironmentError(error_message)


def update_state_file(state_file_path, action_name, action_parent, set_start_time, set_stop_time, result, result_description):
    verify_logger()
    verify_state_file_exists(state_file_path)
    if not action_name or action_name == "":
        error_message = 'action name was not passed'
        logger.exception(error_message)
        raise EnvironmentError(error_message)

    b2b_state = StateDoc(state_file_path)

    b2b_state.ensure_action_exists(action_name)
    if action_parent and action_parent != '':
        logger.debug('Update the parent action')
        b2b_state.update_parent(action_name, action_parent)
    else:
        logger.debug('skip updating the parent action')

    if set_start_time:
        logger.debug('Entering set action start time')
        b2b_state.update_start_time(action_name)
    else:
        logger.debug('skip set action start time')

    if set_stop_time:
        logger.debug('Entering set action stop time')
        b2b_state.update_stop_time(action_name)
    else:
        logger.debug('skip set action start time')

    if result is not None:
        logger.debug('Entering set result')
        b2b_state.update_result(action_name, result)
    else:
        logger.debug('skip set result')

    if result_description:
        logger.debug('Entering set result description')
        b2b_state.update_result_description(action_name, result_description)
    else:
        logger.debug('Skip set result description')

    b2b_state.write_state_doc_actions_update()
    logger.debug('Completed updating history file')


def write_db_sizes(state_file_path, db_phase, sizes):
    if not logger:
        raise NameError('logger is not set')
    if not state_file_path or state_file_path == "":
        error_message = 'State file was not passed'
        logger.exception(error_message)
        raise EnvironmentError(error_message)

    if not db_phase or db_phase == "":
        error_message = 'DB Phase not passed'
        logger.exception(error_message)
        raise EnvironmentError(error_message)

    if not isinstance(sizes, dict) or not len(sizes):
        error_message = 'Must pass valid dictionary with at least one db to update'
        logger.exception(error_message)
        raise EnvironmentError(error_message)

    b2b_state = StateDoc(state_file_path)

    b2b_state.update_db_phase_sizes(db_phase, sizes)


def return_running_action(state_file_path):
    pass


def return_current_action(state_file_path):
    verify_logger()
    verify_state_file_exists(state_file_path)
    b2b_state = StateDoc(state_file_path)
    return b2b_state.return_current_action()


def return_current_action_with_verification(state_file_path):
    """
    Returns the current action, but if the current action has a stop time and isn't apply_system_update, with retries and sleeping inbetween
    """
    retries_remaining = 2
    current_action = None
    while retries_remaining > 0:
        current_action = return_current_action(state_file_path)
        if current_action.stop_time and current_action.action_name != 'b2b_cleanup':
            logger.debug('The action: %s has a stop time (%d), therefore it has completed and it is not the expected last completed step: "apply_system_update", sleep and try again', current_action.action_name, current_action.stop_time)
            retries_remaining -= 1
            time.sleep(5)
        else:
            return current_action
    logger.warning('Failed to find a current action running, check for errors, returning an action if possible')
    return current_action


def return_actions_with_errors(state_file_path):
    """
    Checks all the actions to determine if any actions have a result and that result isn't 0
    """
    verify_logger()
    verify_state_file_exists(state_file_path)
    logger.debug('Entering check for action failures')
    actions_failed = []
    b2b_state = StateDoc(state_file_path)
    for action in b2b_state.action_list:
        if action.result and action.result != 0:
            actions_failed.append(action)
            logger.error('Found failed action: %s, exit code: %d', action.action_name, action.result)
    logger.debug('Returning action failure count: %d', len(actions_failed))
    return actions_failed


if __name__ == "__main__":
    main()

