# !/usr/bin/env python
# Copyright 2015 VMware, Inc.  All rights reserved. -- VMware Confidential
# vROps Product Update Validation Script


__author__ = "VMware, Inc."

import logging
import logging.handlers
import re
import platform
import os
import sys
import subprocess
import time
import shlex
import ctypes
import json
import multiprocessing
import vcops_pak_six
import time
import ssl
import base64
import datetime
import shutil
import glob
import spwd  # Shadow password database (to read /etc/shadow).
import vcops_pack_utilities
import vcops_pack_http_utilities
from vcops_crypt import Crypt
from enum import Enum

STORAGE_CORE_DISK_SPACE_REQUIRED_KEY = 'storage_core_disk_space_required'
STORAGE_LOG_DISK_SPACE_REQUIRED_KEY = 'storage_log_disk_space_required'
STORAGE_DB_DISK_SPACE_REQUIRED_KEY = 'storage_db_disk_space_required'
BINARY_DISK_SPACE_REQUIRED_KEY = 'binaries_disk_space_required'
MANIFEST_FILE_NAME = 'manifest.txt'
PAK_VALIDATE_RESULT_KEY = 'validate_result'
PAK_VALIDATE_EXIT_CODE_KEY = 'validate_exit_code'
PAK_RESOURCE_MESSAGE_KEY = 'resource_message_key'
PAK_RESOURCE_ARGS = 'resource_arguments'
PAK_RESOURCE_DIRECTORY = 'resource_directory'
PAK_VALIDATE_STATUS_FILE_KEY = 'validate_status_file'

STARTING_DIR = os.getcwd()
WARNING_RESULT_TYPE = "WARN"

token_pair = None

class EnvironmentInfo(object):
    def __init__(self):
        pass

    slice_info_list = []
    SINGLE_NODE_MAX_RESOURCE_COUNT = 45000
    MULTI_NODE_MAX_RESOURCE_COUNT = 40000
    MULTI_NODE_MAX_PRIMARY_RESOURCE_COUNT = 20000

    REMOTE_COLLECTOR_ROLE = 'REMOTE_COLLECTOR'
    ONE_WAY_REMOTE_COLLECTOR_ROLE = 'ONE_WAY_REMOTE_COLLECTOR'
    MAX_LATENCY_RC_NODE_MS = 200
    MAX_LATENCY_DATA_NODE_MS = 5
    SWAP_LIMIT = 0.5
    TIME_IN_MINUTES_TO_CHECK = 15
    cluster_online = None
    bypass_prechecks = None

    MINIMUM_ACCEPTABLE_VERSION_INT = 661
    MINIMUM_ACCEPTABLE_VERSION_STR = "6.6.1"
    INITIAL_PHOTON_VERSION_INT = 800
    INITIAL_TOKEN_BASED_AUTH_VERSION_INT = 810

    @classmethod
    def newInstance(cls):
        item = Linux()
        return item

    def run_script(self, scriptWithArgs, acceptableReturnCodes=None, suppressErrors=False, runningDirectory=None,
                   waitForExit=True, useShell=True, universalNewlines=True):
        """
        Runs a script
        :param scriptWithArgs the script and arguments to run
        :param acceptableReturnCodes a list of return codes acceptable upon completion of running the script. Default is [0].
        :return tuple (boolean-script ran successfully, int-exit code, str-stderr, str-stdout)
        """
        scriptCommand = scriptWithArgs
        logger.debug('Script command: "%s"', scriptWithArgs)
        try:
            processPipe = subprocess.Popen(scriptCommand, shell=useShell, cwd=runningDirectory, stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE, universal_newlines=universalNewlines)
        except OSError as e:  # WindowsError may be thrown; can get all info from OSError
            message = 'Caught OSError executing command "%s". Exception "%s"'
            if suppressErrors:
                logger.info(message, scriptWithArgs, e)
            else:
                logger.error(message, scriptWithArgs, e)

            # If there was an errno and it was an acceptable return code, we'll still say we succeeded.
            succeeded = OSError.errno and OSError.errno in acceptableReturnCodes
            return succeeded, e.errno, e, ''

        if waitForExit:
            (processStdOut, processStdErr) = processPipe.communicate()
            logger.debug('stdout: "%s"', processStdOut)
            logger.debug('stderr: "%s', processStdErr)
            logger.debug('exit code: %d', processPipe.returncode)
            if not acceptableReturnCodes:
                acceptableReturnCodes = [0]
            if processPipe.returncode not in acceptableReturnCodes:
                message = 'Script command: "{0}" failed with exit code: {1}'.format(scriptCommand,
                                                                                    processPipe.returncode)
                if suppressErrors:
                    logger.debug(message)
                else:
                    logger.error(message)
                return False, processPipe.returncode, processStdErr, processStdOut
            return True, processPipe.returncode, processStdErr, processStdOut
        else:
            logger.debug('Do not wait for script: "%s" to complete', scriptWithArgs)

    def is_pattern_in_file(self, file_path, search_pattern):
        """
        Checks if a pattern exists in a file
        """
        search_pattern_log_message = "the search pattern: '{0}' in: '{1}'".format(search_pattern, file_path)
        logger.debug("Checking for {0}".format(search_pattern_log_message))

        if not os.path.isfile(file_path):
            logger.error('The file: "{0}" does not exist'.format(file_path))

        # Search for text
        try:
            with open(file_path) as working_file:
                for line in working_file:
                    if re.search(search_pattern, line):
                        logger.info('Found {0}'.format(search_pattern_log_message))
                        return True
        except IOError as e:
            logger.exception('Failed to open the file "{0}"'.format(file_path))
            return False
        logger.info('Failed to find {0}'.format(search_pattern_log_message))
        return False

    def check_for_bad_locator_address(self):
        """
        Checks the _orig version of the gemfire.properties file if localhost or 127.0.0.1 is not in the gemfire.properties
        file, fail.  After the product is configured it should have a real IP or host name here, but this is the _orig
        file that should have the default value.  If we find a non-default value fail as a 6.0.3 HF wrote a bad value
        to the file that will cause the upgrade to fail
        """

        result = True
        result_description = []
        try:
            orig_gemfire_props_file = os.path.join(os.environ['ALIVE_BASE'], '_originalconf', 'gemfire.properties')
        except KeyError as e:
            logger.warning('Failed to set the orig gemfire props file path: {0} as ALIVE_BASE is not set'.format(e))
            return True

        if not os.path.isfile(orig_gemfire_props_file):
            logger.warning(
                'The orig gemfire props file: "{0}" does not exist, skip check'.format(orig_gemfire_props_file))
            return True

        valid_original_values = ['localhost', '127.0.0.1']
        pattern = r"locators\s*=\s*{0}:".format("|".join([re.escape(value) for value in valid_original_values]))
        if not self.is_pattern_in_file(orig_gemfire_props_file, pattern):
            logger.error(
                'The original gemfire.properties file: "{0}" is in a bad state and needs to be corrected.  Search for a KB on this.  KB: TBD'.format(
                    orig_gemfire_props_file))
            result = False
            result_description = [{'key': 'gemfire_locator_bad_state.description'}]

        logger.info('The original gemfire.properties file does not contain the bad strings, check passed!')
        return result, result_description

    def file_text_search(self, file_path, search_pattern):
        """
        Returns a match object for the search pattern on the file
        """
        if not os.path.exists(file_path):
            logger.error('File: "%s" does not exist', file_path)
            return None
        try:
            workingFile = open(file_path, 'r')
            for line in workingFile:
                textFound = re.search(search_pattern, line.rstrip())
                if textFound:
                    logger.info('Text found: "%s" in: "%s"', textFound.group(0), file_path)
                    return textFound
        except IOError:
            logger.exception('Failed to open the file: "%s"', file_path)
            return None
        logger.info('Text NOT found: "%s" in: "%s"', search_pattern, file_path)
        return None

    def load_json_document(self, json_file):
        """
        Reads a Json document and returns that object so it can be read
        @param json_file the Json file to load
        @return object containing Json document
        """
        if not os.path.isfile(json_file):
            logger.error('Json file: "%s" does not exist, cannot load Json document', json_file)
            return None
        else:
            logger.debug('Loaded Json file: "%s"', json_file)
        try:
            with open(json_file) as fp:
                data = fp.read()
                json_document = json.loads(data)  # will work with or without BOM
        except IOError:
            logger.exception('Failed to read the Json file: "%s"', json_file)
            return None
        except ValueError:
            logger.exception('Failed to parse Json file: "%s"', json_file)
            return None
        return json_document

    def load_manifest_value(self, manifest_file, key_name, manifest_document=None, default_value=None):
        """
        Returns a value for a specific key
        """
        logger.debug('Entering load manifest value for: %s', key_name)
        if manifest_document is None:
            logger.debug('The manifest file has not been loaded, load it now')
            manifest_document = self.load_json_document(manifest_file)
        try:
            key_value = manifest_document[key_name]
            logger.debug('Value for: %s is: "%s"', key_name, key_value)
            if key_value is None:
                logger.debug('Returning default key value: "%s"', default_value)
                return default_value
            else:
                logger.debug('Returning key value: "%s"', key_value)
                return key_value
        except KeyError:
            if default_value is not None:
                logger.debug('Returning default key value: "%s"', default_value)
                return default_value
            logger.exception('The key: "%s" does not exist', key_name)
            return None

    def return_dir_size(self, dir_to_check):
        """
        Returns the total directory size
        """
        dir_size = 0
        for (path, dirs, files) in os.walk(dir_to_check):
            for found_file in files:
                filename = os.path.join(path, found_file)
                try:
                    dir_size += os.path.getsize(filename)
                except EnvironmentError:
                    logger.exception('Failed to get the dir size for: "%s"', dir_to_check)
        logger.debug('Dir: "%s" has a total size of: %d', dir_to_check, dir_size)
        return dir_size

    def convert_mega_bytes_to_bytes(self, mega_bytes):
        return mega_bytes * 1024 * 1024

    def convert_giga_bytes_to_bytes(self, giga_bytes):
        return self.convert_mega_bytes_to_bytes(giga_bytes * 1024)

    def convert_bytes_to_giga_bytes(self, bytes):
        return bytes / 1024 / 1024 / 1024

    def get_cpu_count(self):
        return multiprocessing.cpu_count()

    def check_disk_space_needed(self):
        result = True
        result_description = []
        """
        Checks the total space needed for the partition where $STORAGE/db is located
        Returns True if there is enough space of False if there is not
        """
        manifest_file = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), MANIFEST_FILE_NAME)
        manifest_document = self.load_json_document(manifest_file)
        db_partition_size_required = self.convert_mega_bytes_to_bytes(
            self.load_manifest_value(manifest_file, STORAGE_DB_DISK_SPACE_REQUIRED_KEY, manifest_document, 0))
        db_partition_location = VcOpsPaths.find_storage_db_path()
        logger.debug('DB partition size: %d, location: %s', db_partition_size_required, db_partition_location)
        total_storage_required = db_partition_size_required

        partitions_to_check = {VcOpsPaths.find_storage_log_path(): self.convert_mega_bytes_to_bytes(
            self.load_manifest_value(manifest_file, STORAGE_LOG_DISK_SPACE_REQUIRED_KEY, manifest_document, 0)),
                               VcOpsPaths.find_storage_core_path(): self.convert_mega_bytes_to_bytes(
                                   self.load_manifest_value(manifest_file, STORAGE_CORE_DISK_SPACE_REQUIRED_KEY,
                                                            manifest_document, 0)),
                               VcOpsPaths.find_vcops_base_path(): self.convert_mega_bytes_to_bytes(
                                   self.load_manifest_value(manifest_file, BINARY_DISK_SPACE_REQUIRED_KEY,
                                                            manifest_document, 0))}

        for path, required_size in vcops_pak_six.vcops_items(partitions_to_check):
            logger.debug('Checking path: %s with required size: %d', path, required_size)
            # TODO: Determine if they are on the same or different partitions.  For now assume everything is the same parition, we will require a little more disk, but it is safe
            logger.debug('Adding the required size of: %d to the total required as this is on the same partition as DB',
                         required_size)
            total_storage_required += required_size

        logger.info('Total space required: %d', total_storage_required)
        diskFreeSpace = self.return_free_space(db_partition_location)
        logger.debug('disk space required: %d bytes for path "%s", free disk space: %d bytes',
                     total_storage_required, db_partition_location, diskFreeSpace)
        if diskFreeSpace > total_storage_required:
            logger.info('Disk space check for: %s PASSED', db_partition_location)
        else:
            logger.error(
                'DISK SPACE CHECK FAILED- The disk space required: %d bytes (GB: %d), is too big to fit on the disk (or partition): "%s",'
                ' free space: %d bytes (GB: %d).', total_storage_required,
                self.convert_bytes_to_giga_bytes(total_storage_required), db_partition_location, diskFreeSpace,
                self.convert_bytes_to_giga_bytes(diskFreeSpace))
            result = False
            result_description = [{'key': 'check_disk_space_needed_failed.description',
                                   'params': [str(diskFreeSpace), str(total_storage_required), db_partition_location]}]
        return result, result_description

    def is_cluster_online(self):
        """
        Checks if the cluster is in ONLINE state
        :return: Boolean
        """
        if self.cluster_online is None:
            ACCEPTABLE_STATE = 'ONLINE'
            url = '/sysadmin/cluster/status'
            casa_url = vcops_pack_http_utilities.create_casa_url(url)
            data = CasaManager().get_casa_call(casa_url, logger)
            if not data:
                raise NoDataException(casa_url)

            cluster_state = data['cluster_state']
            if cluster_state != ACCEPTABLE_STATE:
                logger.error('Cluster is in {0} state'.format(cluster_state))
                self.cluster_online = False
            else:
                logger.debug('Cluster in Online')
                self.cluster_online = True

        return self.cluster_online

    def get_slice_info(self):
        """
        Get cluster info and build a list with slice info(contains address and uuid)
        :return: list of SliceInfo
        """
        if self.slice_info_list:
            return self.slice_info_list

        url = '/deployment/cluster/current'
        casa_url = vcops_pack_http_utilities.create_casa_url(url)
        data = CasaManager().get_casa_call(casa_url, logger)
        if not data:
            raise NoDataException(casa_url)

        for slice in data:
            slice_roles = self.get_slice_roles(uuid=slice['slice_uuid'])
            slice_info = SliceInfo(slice['slice_address'], slice['slice_uuid'], slice_roles)
            self.slice_info_list.append(slice_info)

        for slice in self.get_slice_info():
            logger.debug(
                'hostname: {0}, uuid: {1}, roles: {2}'.format(slice.slice_address, slice.slice_uuid, slice.slice_roles))
        return self.slice_info_list

    def ping_data_nodes_check(self):
        """
        Checks if data all nodes are reachable
        :return: Boolean with a message
        """
        result = True
        result_description = [{'key': 'ping_data_nodes_check.description'}]

        for slice in self.get_slice_info():
            if self.REMOTE_COLLECTOR_ROLE in slice.slice_roles:
                continue
            hostname = slice.slice_address
            result, result_description = self.ping_node(hostname)
            if not result:
                break

        return result, result_description

    def ping_rc_nodes_check(self):
        """
        Checks if remote collector all nodes are reachable
        :return: Boolean with a message
        """
        result = True
        result_description = [{'key': 'ping_rc_nodes_check.description'}]

        this_hostname = self.get_slice_address()
        for slice in self.get_slice_info():
            slice_address = slice.slice_address
            ## Multi-cloud remote collector node is considering as remote collector node
            ## but check it's accessibility via master node i.e. this node : as there is no other way ping to it
            ## So considering this is master node ping just get the multi-cloud remote collector some response by uuid
            ## If we will get something sensible then the node is alive.
            if EnvironmentInfo.ONE_WAY_REMOTE_COLLECTOR_ROLE in slice.slice_roles:
                slice_uuid = slice.slice_uuid
                logger.debug('Pinging multi-cloud remote collector {0}'.format(slice_address))
                result, result_description = self.ping_mc_rc_node(slice_uuid, this_hostname)
            elif self.REMOTE_COLLECTOR_ROLE in slice.slice_roles:
                logger.debug('Pinging cluster remote collector {0}'.format(slice_address))
                result, result_description = self.ping_node(slice_address)
            ## Stop processing if any ping result fails.
            if not result:
                break

        return result, result_description

    def ping_node(self, hostname):
        result = True
        result_description = []

        url = '/security/ping'
        casa_ping_url = vcops_pack_http_utilities.create_casa_url(url, hostname)

        try:
            CasaManager().get_casa_call(casa_ping_url, logger)
        except Exception as general_exception:
            result = False
            result_description = result_description = [
                {'key': 'ping_nodes_check_failed.description', 'params': [hostname, str(general_exception)]}]
        return result, result_description

    def ping_mc_rc_node(self, slice_uuid, this_hostname):
        """
        Ping the multi-cloud RC node through master node by it's uuid
        Just get the OK result from below URL to see if the node is alive
        """
        result = True
        result_description = []
        casa_ping_url = '/view/node/' + slice_uuid
        try:
            CasaManager().get_casa_call(casa_ping_url, logger, this_hostname)
        except Exception as general_exception:
            result = False
            result_description = [{'key': 'ping_mc_rc_node', 'parrams': [this_hostname, general_exception]}]
        return result, result_description

    def time_sync_check(self):
        """
        Checks if nodes are in time sync by comapring the timestamps of nodes in milliseconds
        and if they cross tolerance seconds return False
        :return: Boolean with a message
        """
        result = True
        result_description = [{'key' : 'time_sync_check.description'}]

        # Get current timestamp of all the nodes
        url = '/sysadmin/slice/ntp/current_time'

        timestamp_dict = {}
        timestamps = []
        tolerance_seconds = 30

        start_time = datetime.datetime.now()
        for slice in self.get_slice_info():
            hostname = slice.slice_address
            casa_time_sync_url = vcops_pack_http_utilities.create_casa_url(url, hostname)
            ## Skip time synchronisation check for multi-cloud remote collectors (currently this fails the pre-upgrade-check)
            ## Cloud remote collectors can not reach the cluster NTP servers and why the product update should fail because
            ## of big-time difference between cluster and MC RC nodes ?
            ## TODO - "RC design optimization" sets to support "Custom NTP" servers configuration for cloud nodes w.r.t them
            ## we might change this check later.
            if EnvironmentInfo.ONE_WAY_REMOTE_COLLECTOR_ROLE in slice.slice_roles:
                logger.debug('skipping time sync check for multi-cloud rc node {0}'.format(hostname))
                continue
            time_data = CasaManager().get_casa_call(casa_time_sync_url, logger)
            if not time_data:
                raise NoDataException(casa_time_sync_url)

            timestamp = time_data['currentTime']
            time_now = datetime.datetime.now()
            timestamp -= int((time_now - start_time).total_seconds()*1000)
            timestamps.append(timestamp)
            timestamp_dict[timestamp] = hostname
            logger.debug('slice {0} --- timestamp {1}'.format(hostname, timestamp))

        # Sort all timestamps and compare the difference between first and the last
        timestamps = sorted(timestamps)
        time_diff = timestamps[-1] - timestamps[0]
        if time_diff > tolerance_seconds*1000:
            result = False
            result_description = [{'key' : 'time_sync_check_failed.description', 'params' : [timestamp_dict.get(timestamps[0]), timestamp_dict.get(timestamps[-1]), time_diff/1000]}]

        return result, result_description

    def getCasaManager(self):
        if EnvironmentInfo.get_product_version() >= EnvironmentInfo.INITIAL_TOKEN_BASED_AUTH_VERSION_INT:
            return CasaManager("token")
        return CasaManager()

    def self_monitor_check(self):
        """
        Checks if there are any self monitoring alerts in the system
        :return: Boolean with a message
        """
        result = True
        result_description = []
        alert_tolerance_count = 1

        # Get alerts
        suite_api_url = '/suite-api/api/alerts/query'
        request_data = '{"resource-query":{"adapterKind": ["vCenter Operations Adapter"]},"activeOnly": true}'
        data = CasaManager().post_suite_api_call(suite_api_url, request_data)
        if not data:
            raise NoDataException(suite_api_url)

        alert_count = data['pageInfo']['totalCount']
        if alert_count != 0:
            alertDefinitionNames = []
            if alert_count > alert_tolerance_count:
                result = False
            for alert in data['alerts']:
                if 'alertDefinitionName' in alert:
                    alertDefinitionNames.append(alert['alertDefinitionName'])
                else:
                    alertDefinitionNames.append(alert['alertDefinitionId'])
            result_description = [
                {'key': 'self_monitor_check_failed.description', 'params': [alert_count, alertDefinitionNames]}]
            return result, result_description

        if not result_description:
            result_description = [{'key': 'self_monitor_check.description'}]

        return result, result_description

    def latency_check_rc_nodes(self):
        result = True
        result_description = []

        total = 0
        total_count = 0
        casa_ping_url = '/security/ping'
        authorization_required = False
        for slice in self.get_slice_info():
            roles = slice.slice_roles
            if EnvironmentInfo.REMOTE_COLLECTOR_ROLE not in roles:
                continue
            if EnvironmentInfo.ONE_WAY_REMOTE_COLLECTOR_ROLE in roles:
                ## For the multi-cloud remote collector nodes we have only 1-way HTTPS connection so
                ## use this url to get the ping url.
                mc_rc_uuid = slice.slice_uuid
                casa_ping_url = '/view/node/' + mc_rc_uuid
                authorization_required = True

            try:
                latency = CasaManager().get_latency_for_call(casa_ping_url,
                                                             slice.slice_address,
                                                             authorization_required=authorization_required)
            except Exception as general_exception:
                result = False
                result_description = [
                    {'key': 'pinging_rc_nodes_failed.description', 'params': [slice.slice_address, general_exception]}]
                return result, result_description

            logger.debug('Ping time to slice "{0}" is {1}ms'.format(slice.slice_address, latency))
            total += latency
            total_count += 1

            if latency > self.MAX_LATENCY_RC_NODE_MS:
                result = False
                result_description = [
                    {'key': 'latency_check_rc_nodes_failed', 'params': [slice.slice_address, latency]}]
                break
        if result and total_count != 0:
            result_description = [{'key': 'latency_check_rc_nodes.description', 'params': [total / total_count]}]

        return result, result_description

    def ntp_check(self):
        """
        Checks if NTP is configured with same set of time servers
        :return: Boolean with a message
        """
        result = True
        result_description = [{'key': 'ntp_check.description'}]

        # Get NTP information and see if all the nodes are configured with the same time servers
        url = '/sysadmin/nodes/ntp'
        casa_time_sync_url = vcops_pack_http_utilities.create_casa_url(url)

        ntp_data = CasaManager().get_casa_call(casa_time_sync_url, logger)
        if not ntp_data:
            raise NoDataException(casa_time_sync_url)

        first_list = None
        first_node = None
        for slice in ntp_data:
            slice_address = slice['nodeAddress']
            if EnvironmentInfo.ONE_WAY_REMOTE_COLLECTOR_ROLE in self.get_slice_roles_by_ip(slice_address):
                ## Skip time synchronisation check for multi-cloud remote collectors
                ## Cloud 1-way remote collectors can not reach the cluster NTP servers
                ## TODO - "RC design optimization" sets to support "Custom NTP" servers configuration for cloud nodes w.r.t them
                ## we might change this check later.
                logger.info(
                    'Skipping multi-cloud remote collector slice {} NTP configuration check'.format(slice_address));
                continue;

            node_list = []
            for time_server in slice['ntpConfigDocument']['time_servers']:
                node_list.append(time_server['address'])

            if not first_list:
                first_list = node_list
                first_node = slice_address
            else:
                if not list(set(first_list) & set(node_list)):
                    result = False
                    result_description = [
                        {'key': 'ntp_check_failed.description', 'params': [first_node, slice['nodeAddress']]}]
                    break

        return result, result_description

    def get_ip_address_from_online_state_url(self, online_state_url):
        '''
        method removes the URI and https from the online_state_url
        Ex : "https://10.152.186.71:443/casa/sysadmin/slice/online_state" will return 10.152.186.71
             "https://fd01:0:101:2605:0:cff:fe09:166d:443/casa/sysadmin/slice/online_state" will return fd01:0:101:2605:0:cff:fe09:166d
             "https://prome-1s-dhcp43.eng.vmware.com:443/casa/sysadmin/slice/online_state will return prome-1s-dhcp43.eng.vmware.com

        Args:
            online_state_url: the online_state_url

        Returns:
            ip address
        '''
        return online_state_url.split(":", 1)[1].replace(':443/casa/sysadmin/slice/online_state', '').replace('/', '')

    @staticmethod
    def _return_slice_states():
        url = '/sysadmin/cluster/online_state'
        casa_url = vcops_pack_http_utilities.create_casa_url(url)
        logger.debug('entering return cluster online state doc from url: "%s"', casa_url)
        data = CasaManager().get_casa_call(casa_url, logger)
        if not data:
            raise NoDataException(casa_url)
        slice_online_states_doc = data['slice_online_states']
        logger.debug('returning cluster online state doc: %s', slice_online_states_doc)
        return slice_online_states_doc

    def node_online_check(self):
        """
        Checks if the node is in ONLINE state
        :return: Boolean with a message
        """
        WARN_STATE = 'OFFLINE'
        ERROR_STATES = ['FAILURE', 'UNREACHABLE', 'UNKNOWN', 'REMOVING']

        result = True
        result_description = [{'key': 'node_online_check.description'}]
        check_type = None

        for slice_state in EnvironmentInfo._return_slice_states():
            if len(slice_state['slices']) > 0:
                if slice_state['slice_online_state'] in ERROR_STATES:
                    check_type = 'ERROR'
                    result = False
                    for slice_data in slice_state['slices']:
                        result_description = [{'key': 'node_online_check_failed.description', 'params': [
                            self.get_ip_address_from_online_state_url(slice_data["href"])]}]
                    result_description = [{'key': 'node_online_check_state_failed.description',
                                           'params': [slice_state['slice_online_state']]}]
                elif slice_state['slice_online_state'] == WARN_STATE:
                    if not check_type:
                        check_type = 'WARN'
                        result = False
                        result_description = 'Nodes '
                        for slice_data in slice_state['slices']:
                            result_description = [{'key': 'node_online_check_failed.description', 'params': [
                                self.get_ip_address_from_online_state_url(slice_data["href"])]}]
                        result_description = [{'key': 'node_online_check_state_failed.description',
                                               'params': [slice_state['slice_online_state']]}]
        if check_type == 'WARN':
            # OFFLINE case: Update the json and change the check to warn
            result = False
            state_file = StateDoc.construct_state_file_path()
            checks_data = self.read_from_json_file(state_file)
            for check in checks_data['checks']:
                if check['check_name'] == 'node_online_check':
                    check['check_type'] = 'WARN'
            self.write_to_json_file(checks_data, state_file)

        return result, result_description

    def verify_consistent_node_online_states(self):
        """
        checks to ensure all nodes are in the same state (online, offline, etc.  If they are not raise an error
        this check is just to verify consistency, other checks like node_online_check will verify the specific values
        """
        logger.debug('entering verify consistent node online states check')
        node_state_dict = {}
        for node_state in EnvironmentInfo._return_slice_states():
            node_state_count = len(node_state['slices'])
            if node_state_count > 0:
                node_state_dict[node_state['slice_online_state']] = node_state_count
                logger.debug('found: %d nodes with the online state: "%s"', node_state_count, node_state)
        logger.info('node online state counts: {0}'.format(node_state_dict))
        if len(node_state_dict) > 1:
            result = False
            result_string = 'failed'
        else:
            result = True
            result_string = 'pass'
        result_description = [{'key': 'verify_consistent_node_online_states.description', 'params': [result_string]}]
        logger.info('completed running verify consistent node online states check, result: %s: %s', result_string,
                    node_state_dict)
        return result, result_description

    def is_ha_enabled(self):
        """
        Checks if HA is enabled
        :return: Boolean
        """
        url = '/config/cluster/ha'
        casa_url = vcops_pack_http_utilities.create_casa_url(url)
        data = CasaManager().get_casa_call(casa_url, logger)
        if not data:
            raise NoDataException(casa_url)

        if data['ha_state'] != 'ENABLED':
            return False
        else:
            return True

    def cluster_degraded_check(self):
        """
        Checks if HA is enabled and cluster is in degraded state
        :return: Boolean with a message
        """
        result = True
        result_description = []

        if not self.is_ha_enabled():
            result_description = [{'key': 'ha_is_not_enabled.description'}]
            return result, result_description

        url = '/config/cluster/ha'
        casa_url = vcops_pack_http_utilities.create_casa_url(url)
        data = CasaManager().get_casa_call(casa_url, logger)
        if not data:
            raise NoDataException(casa_url)

        degraded = data['ha_details']['is_ha_degraded']
        if degraded:
            result = False
            result_description = [{'key': 'ha_is_enabled.description'}]

        return result, result_description

    def cluster_sync_check(self):
        """
        Checks if HA is enabled and nodes are in sync state
        :return: Boolean with a message
        """
        result = True
        result_description = []

        if not self.is_ha_enabled():
            result_description = [{'key': 'ha_is_not_enabled.description'}]
            return result, result_description

        url = '/status/sync/cluster'
        casa_url = vcops_pack_http_utilities.create_casa_url(url)
        data = CasaManager().get_casa_call(casa_url, logger)
        if not data:
            result_description = [{'key': 'cluster_sync_check.description'}]
            return result, result_description

        if len(data) != 0:
            result = False
            try:
                for slice_info in data:
                    slice_address = self.get_slice_address_from_uuid(slice_info['slice_uuid'])
                    if slice_address:
                        result_description = [{'key': 'cluster_sync_check_failed.description',
                                               'params': [slice_address, slice_info['percent_complete']]}]
            except Exception as ex:
                logger.error('Exception while generating error message: {0}'.format(ex))
            return result, result_description

    def get_slice_address_from_uuid(self, uuid):
        list = self.get_slice_info()

        for slice_info in list:
            if slice_info.slice_uuid == uuid:
                return slice_info.slice_address
        return None

    ## Given the ip of the slice just search the slice_info_list for matching
    ## and return the matched ip (slice) roles
    def get_slice_roles_by_ip(self, ip):
        list = self.get_slice_info()

        for slice_info in list:
            if ip == slice_info.slice_address:
                return slice_info.slice_roles

        return None

    def dns_node_resolution_check(self):
        """
        /casa/deployment/cluster/current gives us the host names of the nodes
        Try to ping to the host names so that before pinging hostname will be resolved
        :return:
        """
        result = True
        result_description = []

        # No need to do this on master since it is already done
        if self.admin_role_enabled():
            result_description = [{'key': 'skipping_dns_node_resolution_check'}]
            return result, result_description

        if self.REMOTE_COLLECTOR_ROLE in self.get_slice_roles():
            return self.ping_data_nodes_check() and self.ping_rc_nodes_check()
        else:
            return self.ping_data_nodes_check()

    def get_slice_address(self):
        url = '/deployment/slice'
        casa_url = vcops_pack_http_utilities.create_casa_url(url)
        data = CasaManager().get_casa_call(casa_url, logger)
        if not data:
            raise NoDataException(casa_url)

        return data['slice_address']

    def get_slice_uuid(self):
        url = '/deployment/slice'
        casa_url = vcops_pack_http_utilities.create_casa_url(url)
        data = CasaManager().get_casa_call(casa_url, logger)
        if not data:
            raise NoDataException(casa_url)

        return data['slice_uuid']

    def resource_count_check(self):
        """
        Checks if number of resources in the system exceeds certain number
        :return: Boolean with a message
        """
        result = True
        result_description = []
        num_resources = -1

        slice_uuid = self.get_slice_uuid()
        if not slice_uuid:
            result = False
            result_description = [{'key': 'get_slice_uuid_failed.description'}]
            return result, result_description

        suite_api_url = '/suite-api/internal/casa/clusters/info/'
        data = None
        try:
            data = CasaManager().get_suite_api_call(suite_api_url)
        except RestException as re:
            response_code = re.respone.status
            if response_code == 503:
                result_description = [{'key' : 'check_suite_api_state.description'}]
            elif response_code >= 400 and response_code <= 499:
                # Some API might not be implemented in earlier versions of vROps,
                result_description = [{'key' : 'ignore_test_response_code.description', 'params' : [response_code]}]
            return result, result_description
        if not data:
            raise NoDataException(suite_api_url)

        resourceCount = 'localResourceCount'
        if self.is_ha_enabled() == True:
            resourceCount = 'localPrimaryResourceCount'

        node_count = 0
        try:
            for lduInfo in data['lduinfos']:
                analyticsInfos = lduInfo['analyticsInfos']
                for analyticsInfo in analyticsInfos:
                    node_count += 1
                    if slice_uuid == analyticsInfo['instanceId']:
                        num_resources = analyticsInfo[resourceCount]
        except KeyError as ke:
            logger.error('Exception while parsing json {0}'.format(ke))

        if node_count == 0:
            result = False
            result_description = []
            return result, result_description

        if node_count == 1:
            if num_resources > self.SINGLE_NODE_MAX_RESOURCE_COUNT:
                result = False
                result_description = [
                    {'key': 'resource_count_check.description', 'params': [self.get_slice_address(), num_resources]}]
        else:
            if self.is_ha_enabled() == True:
                if num_resources > self.MULTI_NODE_MAX_PRIMARY_RESOURCE_COUNT:
                    result = False
                    result_description = [{'key': 'resource_count_check.description',
                                           'params': [self.get_slice_address(), num_resources]}]
            else:
                if num_resources > self.MULTI_NODE_MAX_RESOURCE_COUNT:
                    result = False
                    result_description = [{'key': 'resource_count_check.description',
                                           'params': [self.get_slice_address(), num_resources]}]

        return result, result_description

    def cluster_rebalance_check(self):
        """
        Checks if the cluster in rebalance state
        :return: Boolean with a message
        """
        result = True
        result_description = []

        url = '/status/rebalance/cluster'
        casa_url = vcops_pack_http_utilities.create_casa_url(url)
        data = CasaManager().get_casa_call(casa_url, logger)
        if not data:
            result_description = [{'key': 'cluster_rebalance_check.description'}]
            return result, result_description

        if len(data) != 0:
            result = False
            result_description = [
                {'key': 'cluster_rebalance_check_failed.description', 'params': [data['percent_complete']]}]

        return result, result_description

    def epops_agent_existence_check(self):
        """
           Checks if the EPOps agent is present
           :return: Boolean with a message
        """

        result = False
        result_description = [{'key': 'epops_agent_existence_check.description'}]
        return result, result_description

    def verify_jump_release_for_cp_renewal(self):
        logger.info('performing cloud proxy jump check')
        result = True
        result_description = []
        release_number = EnvironmentInfo.return_release_number(EnvironmentInfo.return_build_info_file())
        version_split = release_number.split('.')
        major = version_split[0]
        minor = version_split[1]
        if major == "8" and minor in ['4', '5']:
            logger.info("Version is 8.4 or 8.5")
            cp_list = self.get_cloud_proxy_list()
            cp_count = len(cp_list)
            if cp_count:
                logger.error("Upgrade is not permitted")
                result = False
                result_description = [{'key': 'verify_jump_release_for_cp_renewal.description'}]

        return result, result_description

    def get_cloud_proxy_list(self):
        """
        Perform call to "/casa/saas/v1/cloudproxies" to retrieve list of CPs
        """
        logger.info('Making CaSA call to retrieve CP list v1')
        global token_pair
        call_type = 'GET'
        url = '/saas/v1/cloudproxies'
        casa_url = vcops_pack_http_utilities.create_casa_url(url)
        logger.debug('Calling "{0}": "{1}"'.format(call_type, casa_url))

        response = CasaManager().get_casa_call(casa_url, logger)
        if not response:
            raise NoDataException(casa_url)

        resultsJson = response
        cloud_proxies = resultsJson["cloudproxies"]
        return cloud_proxies

    def telegraf_check(self):
        """Checking is there any APPOSUCP adapter for which "CP" identifier value is "false"""

        result = True
        result_description = []
        suite_api_url = '/suite-api/api/adapters'

        data = CasaManager().get_suite_api_call(suite_api_url)

        for adapter_instance in data['adapterInstancesInfoDto']:
            if adapter_instance['resourceKey']['adapterKindKey'] == 'APPOSUCP':
                for resource_identifier in adapter_instance['resourceKey']['resourceIdentifiers']:
                    if resource_identifier['identifierType']['name'] == "CP" and resource_identifier['value'] == 'false':
                        result = False
                        result_description = [{'key': 'telegraf_check.description'}]

        return result, result_description

    def log_insight_adapter_check(self):

        """Checking if LogInsight Adapter is installed"""

        result = True
        result_description = []
        suite_api_url = '/suite-api/api/adapters'

        data = CasaManager().get_suite_api_call(suite_api_url)

        for adapter_instance in data['adapterInstancesInfoDto']:
            if adapter_instance['resourceKey']['adapterKindKey'] == 'LogInsightAdapter':
                result_description = [{'key': 'log_insight_adapter_check.description'}]
                result = False
                break

        return result, result_description

    def management_pak_existence_check(self):

        """Checking if installed Management Pack versions are less than requited"""

        result = True
        result_description = []

        data = envInfo.read_from_json_file(StateDoc.get_mp_file_path())

        if not data:
            logger.error('Couldn\'t get Management Paks from "{0}"'.format(StateDoc.get_mp_file_path()))
            sys.exit(1)

        management_paks = data['management_paks']

        url = '/suite-api/api/solutions'
        solutions = CasaManager().get_suite_api_call(url)
        mp_list_for_current_setup = {}
        return_param = ''

        # getting MP versions from solutions
        for solution in solutions['solution']:
            for mp in management_paks:
                if solution['adapterKindKeys']:
                    if mp['adapterKindKey'] in solution['adapterKindKeys']:
                        mp_list_for_current_setup[mp['adapterKindKey']] = ".".join(solution['version'].split('.')[:3])

        for management_pak in management_paks:
            if management_pak['adapterKindKey'] in mp_list_for_current_setup.keys():
                if mp_list_for_current_setup[management_pak['adapterKindKey']] <= management_pak['version']:
                    return_param += '\n {} / {} / Greater than {}'.format(management_pak['name'],
                                                             mp_list_for_current_setup[management_pak['adapterKindKey']],
                                                             management_pak['version'])
                    result = False
                    result_description = [{'key': 'management_pak_existence_check.description', 'params':[return_param]}]

        return result, result_description

    def service_restarts_check(self):
        """
        Checks if any services got STARTED/STOPPED/RESTARED in the last 3 hrs
        :return:
        """
        result = True
        result_description = []

        service_history_dict = {}
        service_states = {'Starts': 'STARTED',
                          'Stops': 'STOPPED',
                          'Restarts': 'RESTARTED'}

        watchdog_history_path = os.path.join(VcOpsPaths.find_storage_log_path(), 'log', 'vcops-watchdog',
                                             'vcops-watchdogHistory.json')
        try:
            if not os.path.isfile(watchdog_history_path):
                result_description = [{'key': 'watchdog_file_not_exist.description'}]
                return result, result_description
            logger.debug('Reading from "{0}" watchdog state file'.format(watchdog_history_path))
            with open(watchdog_history_path) as w_history:
                data = json.load(w_history)
        except EnvironmentError as e:
            result = False
            result_description = [{'key': 'watchdog_file_read_fail.description', 'params': [watchdog_history_path]}]
            return result, result_description
        if not data:
            result_description = [{'key': 'service_restarts_check.description'}]
            return result, result_description

        logger.debug(json.dumps(data, indent=4, sort_keys=True))
        for item in data['actions']:
            action_time_str = item['date']
            time_format_for_action = "%Y-%m-%d %H:%M:%S.%f" if '.' in action_time_str else "%Y-%m-%d %H:%M:%S"
            action_time = datetime.datetime.strptime(action_time_str, time_format_for_action)
            current_time = datetime.datetime.now()
            time_diff = current_time - action_time
            if time_diff.total_seconds() / 60 < self.TIME_IN_MINUTES_TO_CHECK:
                service = item['service']
                action = service_states.get(item['action'])
                if not action:
                    action = item['action']

                if service in service_history_dict:
                    action_dict = service_history_dict.get(service, {})
                    if action in action_dict:
                        action_dict[action] = action_dict.get(action) + 1
                    else:
                        action_dict[action] = 1
                    service_history_dict[service] = action_dict
                else:
                    action_dict = {action: 1}
                    service_history_dict[service] = action_dict

        for key, value in vcops_pak_six.vcops_items(service_history_dict):
            result_description.append({'key': 'service_restarts_check_service.description', 'params': [key.upper()]})
            for key2, value2 in vcops_pak_six.vcops_items(value):
                result_description.append({'key': 'service_restarts_check_times.description', 'params': [key2, value2]})

        if service_history_dict:
            result_description.append(
                {'key': 'service_restarts_check_last_minutes.description', 'params': [self.TIME_IN_MINUTES_TO_CHECK]})
            result = False
        else:
            result_description = [{'key': 'service_restarts_check.description'}]

        return result, result_description

    def bootstrap_file_existence_check(self):
        result = True
        result_description = []
        if os.path.isfile("/storage/db/vcops/recovery.conf.bootstrap"):
            result = False
            result_description = [{'key': 'bootstrap_file_existence_check.failed.description'}]
        return result, result_description

    def update_state_file(self, check_name, set_start_time=False, set_stop_time=False,
                          result=None, result_description=None, new_result_type=None):
        if not check_name or check_name == "":
            error_message = 'check name was not passed'
            logger.exception(error_message)
            raise EnvironmentError(error_message)

        logger.debug('Updating state file with check: {0}, start_time: {1}, stop_time: {2}, '
                     'result: {3}, result_desc: {4}, new_result_type: {5}'
                     .format(check_name, set_start_time, set_stop_time, result, result_description, new_result_type))
        b2b_state = StateDoc()

        if set_start_time:
            b2b_state.update_start_time(check_name)

        if set_stop_time:
            b2b_state.update_stop_time(check_name)

        if result is not None:
            b2b_state.update_result(check_name, result)

        if result_description:
            b2b_state.update_result_description(check_name, result_description)

        if new_result_type:
            b2b_state.update_result_type(check_name, new_result_type)

        b2b_state.write_state_doc_actions_update()

    def read_from_json_file(self, state_file_path):
        logger.debug('Reading from state file: "{0}"'.format(state_file_path))
        if os.path.isfile(state_file_path):
            with open(state_file_path) as state_file:
                data = json.load(state_file)
                return data
        else:
            logger.error('State file "{0}" doesn\'t exist'.format(state_file_path))
            return None

    def write_to_json_file(self, jsonDocument, jsonFile, sortKeys=True):
        logger.debug('Writing to json file "{0}"'.format(jsonFile))
        try:
            with open(jsonFile, 'w') as workingJsonFile:
                json.dump(jsonDocument, workingJsonFile, indent=4, sort_keys=sortKeys)
        except EnvironmentError as e:
            logger.exception('Failed to write to the json file: "{0}"'.format(jsonFile))

    def get_check(self, check_name):
        checks_data = envInfo.read_from_json_file(StateDoc.construct_state_file_path())
        if not checks_data:
            logger.error('Couldn\'t read from "{0}"'.format(StateDoc.construct_state_file_path()))
            return None
        for pre_check in checks_data['checks']:
            if pre_check['check_name'] == check_name:
                return pre_check

    def copy_bypass_checks(self, pak_root):
        pre_checks_json = 'bypass_prechecks.json'
        pak_repo_local = os.path.dirname(pak_root)
        old_path = os.path.join(pak_root, 'extracted', pre_checks_json)
        new_path = self.get_bypass_checks_file_location(pak_root)
        if not os.path.isfile(new_path):
            shutil.copyfile(old_path, new_path)

    def bypass_preupgrade_checks(self, pak_root):
        if self.bypass_prechecks is None:
            bypass_json_path = self.get_bypass_checks_file_location(pak_root)
            data = self.read_from_json_file(bypass_json_path)
            if not data:
                logger.error('Couldn\'t read from bypass checks file')
                self.bypass_prechecks = False
                return self.bypass_prechecks

            if str(data['bypass_preupgrade_checks']).lower() == 'true'.lower():
                logger.info('Skipping pre-upgrade checks since bypass is set to TRUE at "{0}"'.format(bypass_json_path))
                self.bypass_prechecks = True
                return self.bypass_prechecks
            else:
                self.bypass_prechecks = False
                return self.bypass_prechecks

        logger.info('Bypass preupgrade checks is set to {0}'.format(self.bypass_prechecks))
        return self.bypass_prechecks

    def get_bypass_checks_file_location(self, pak_root):
        pak_id = os.path.basename(pak_root)
        pak_repo_local = os.path.dirname(pak_root)
        return os.path.join(pak_repo_local, 'bypass_prechecks_{0}.json'.format(pak_id))

    def update_pak_results_file(self, exit_code=None, failed_check=None, message=None, directory=None, args=None,
                                state_file=False):
        pak_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        results_file = self.get_pak_results_file(pak_root)
        logger.debug("results_file :{0}".format(results_file))
        if not os.path.isfile(results_file):
            logger.error('Results file "{0}" doesn\'t exist'.format(results_file))
            return
        results_file_bak = '{0}.bak'.format(results_file)
        shutil.move(results_file, results_file_bak)
        results_data = self.read_from_json_file(results_file_bak)
        if results_data:
            if exit_code:
                results_data[PAK_VALIDATE_EXIT_CODE_KEY] = exit_code
            if failed_check:
                results_data[PAK_VALIDATE_RESULT_KEY] = failed_check
            if message:
                results_data[PAK_RESOURCE_MESSAGE_KEY] = message
            if directory:
                results_data[PAK_RESOURCE_DIRECTORY] = directory
            if args:
                results_data[PAK_RESOURCE_ARGS] = args
            if state_file:
                results_data[PAK_VALIDATE_STATUS_FILE_KEY] = StateDoc.construct_state_file_path()

            logger.debug('Updating results file with exit_code: {0}, exit_message: {1}, res_message: {2},' \
                         ' res_directory: {3}, res_args: {4}, res_state_file: {5}' \
                         .format(exit_code, failed_check, message, directory, args, state_file))
            self.write_to_json_file(results_data, results_file_bak)
            shutil.move(results_file_bak, results_file)

    def get_pak_results_file(self, pak_root):
        pak_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        pak_id = os.path.basename(pak_root)
        results_file = os.path.join(pak_root, '{0}.results'.format(pak_id))
        logger.debug('Results file: "{0}"'.format(results_file))
        return results_file

    def get_slice_roles(self, uuid=None):
        result = []
        url = '/deployment/slice/role/status'
        casa_url = vcops_pack_http_utilities.create_casa_url(url)
        if uuid:
            casa_url = casa_url + '/' + uuid
        data = CasaManager().get_casa_call(casa_url, logger)

        for key, value in vcops_pak_six.vcops_items(data['enabled']):
            if value:
                result.append(key)
        slice_roles = result

        return slice_roles

    def admin_role_enabled(self):
        list = self.get_slice_roles()
        if 'ADMIN' in list and 'REPLICA' not in list:
            return True
        else:
            return False

    def data_role_enabled(self):
        list = self.get_slice_roles()
        if 'DATA' in list:
            return True
        else:
            return False

    def admin_or_replica_role_enabled(self):
        list = self.get_slice_roles()
        if 'ADMIN' in list:
            return True
        else:
            return False

    def get_partitions_for_path(self, path):
        raise NotImplementedError

    def return_free_space(self, directory):
        raise NotImplementedError

    def return_pak_id(self):
        logger.debug('Setting current dir to the starting dir: "%s"', STARTING_DIR)
        try:
            os.chdir(STARTING_DIR)
        except EnvironmentError:
            logger.exception('Failed to switch directory to %s', STARTING_DIR)
        pak_id = os.path.basename(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        logger.debug('Pak ID: "%s"', pak_id)
        if not pak_id:
            logger.error('Failed to find a pak ID, return None')
        return pak_id

    def set_pak_action_failure(self, exit_code, exit_reason, resource_message_key, resource_message_args,
                               resource_directory=None):
        logger.debug('Attempting to set pak failure for the current pak, exit code: %s, exit reason: %s, resource '
                     'message key: %s, resource message args: %s, resource directory: %s', exit_code, exit_reason,
                     resource_message_key, ''.join(resource_message_args), resource_directory)

        pak_manager_py = os.path.join(os.path.dirname(VcOpsPaths.find_vcops_base_path()), 'vmware-vcopssuite',
                                      'utilities', 'pakManager', 'bin', 'vcopsPakManager.py')
        if not os.path.isfile(pak_manager_py):
            logger.warning('Pak Manager file: "%s" does not exist, cannot update action failure', pak_manager_py)
            return

        pak_id = self.return_pak_id()
        if not pak_id:
            logger.warning('Cannot find a pak id, cannot update action failure')
            return

        # check to ensure we are running on a version of vROps Pak Manager that support setting the failure, if
        # we aren't, return
        pak_manager_set_script_failure_action = 'set_script_failure'
        pak_manager_help_result = self.run_script('{0} {1} -h'.format(sys.executable, pak_manager_py),
                                                  acceptableReturnCodes=[1])
        if pak_manager_set_script_failure_action not in pak_manager_help_result[3]:
            logger.warning('Setting the script failure ("%s") is not supported in the current version of Pak Manager, '
                           'skip it', pak_manager_set_script_failure_action)
            return

        # run set script failure
        validate_script_failure = '{0} {1} --action {2} --pak {3} --exit_code {4} --exit_message "{5}" ' \
                                  '--action_to_update validate'.format(sys.executable, pak_manager_py,
                                                                       pak_manager_set_script_failure_action, pak_id,
                                                                       exit_code, exit_reason)
        if resource_message_key:
            validate_script_failure += ' --resource_message_key "{0}"'.format(resource_message_key)
        if resource_message_args and len(resource_message_args) > 0:
            for arg in resource_message_args:
                validate_script_failure += ' --resource_message_args "{0}"'.format(arg)
        if resource_directory:
            validate_script_failure += ' --resource_directory ' + resource_directory
        logger.debug('Calling pak manager set action failure, command: "%s"', validate_script_failure)
        if not self.run_script(validate_script_failure)[0]:
            logger.warning('Failed calling pak manager set failure (command: "%s")', validate_script_failure)

    def load_dictionary_key_value(self, dictionary_file, key):
        key_string = key + '\s*='
        re_search_pattern = r'^' + key_string + '.*$'
        if not os.path.exists(dictionary_file):
            raise RuntimeError('File: "{0}" does not exist'.format(dictionary_file))
        key_returned_string = None
        try:
            with open(dictionary_file, 'r') as working_file:
                for line in working_file:
                    key_returned_string = re.search(re_search_pattern, line)
                    if key_returned_string:
                        break
        except EnvironmentError as e:
            raise RuntimeError(
                'Failed to open the file: "{0}" caused by exception: {1}'.format(dictionary_file, str(e)))
        if key_returned_string:
            value_string = key_returned_string.group(0).split('=', 1)[1].strip()
            value_string = value_string.replace("\\", "")
            return value_string
        else:
            raise RuntimeError('Failed to find the key: "{0}" in the file: "{1}"'.format(key, dictionary_file))

    @staticmethod
    def return_release_number(build_info_file=None):
        logger.debug('entering return release number')
        release_number_to_return = None
        if build_info_file:
            logger.debug('get release number from: "%s"', build_info_file)
            try:
                with open(build_info_file) as f:
                    for line in f.readlines():
                        if line.startswith('RELEASENUMBER:'):
                            release_number_to_return = line.split(':')[1]
            except EnvironmentError:
                logger.exception('Error occurred reading file "%s"', build_info_file)
        else:
            logger.error('build info file not provided')
        logger.debug('returning release number: "%s"', release_number_to_return)
        return release_number_to_return

    @staticmethod
    def get_product_version(humanReadable=False):
        logger.debug('entering get_product_version')
        build_info_file = EnvironmentInfo.return_build_info_file()
        release_number_full = EnvironmentInfo.return_release_number(build_info_file)
        product_version = None
        if release_number_full:
            logger.debug('get product version from release number: "%s"', release_number_full)
            try:
                product_version = release_number_full.rsplit('.', 1)[0]
                if not humanReadable:
                    product_version = int(product_version.replace('.', ''))
            except:
                logger.error('could not get product version from release number: "%s"', release_number_full)
        else:
            logger.error('could not get release number from build_info_file: "%s"', build_info_file)
        logger.debug('returning product version: "%s"', product_version)
        return product_version

    @staticmethod
    def return_build_info_file():
        logger.debug('return build info file')
        build_info_file = os.path.join(os.path.dirname(VcOpsPaths.find_vcops_base_path()),
                                       'vmware-vcopssuite-installsupport',
                                       '.buildInfo')
        logger.debug('return build info file: "%s"', build_info_file)
        return build_info_file

    @staticmethod
    def version_check():
        logger.debug('Entering version check')
        currentVersion = EnvironmentInfo.get_product_version()
        currentVersionStr = EnvironmentInfo.get_product_version(humanReadable=True)
        if currentVersion:
            logger.debug("Product's current version is: {0}".format(str(currentVersion)))
            if currentVersion < EnvironmentInfo.MINIMUM_ACCEPTABLE_VERSION_INT:
                logger.debug("Product's current version:{0} "
                             "is less then minimum acceptable version:{1}".format(currentVersionStr,
                                                                                  EnvironmentInfo.MINIMUM_ACCEPTABLE_VERSION_STR))
                return False, [{'key': 'minimum_version_check_failed.description',
                                'params': [currentVersionStr,
                                           EnvironmentInfo.MINIMUM_ACCEPTABLE_VERSION_STR]}]
            if currentVersion < EnvironmentInfo.INITIAL_PHOTON_VERSION_INT:
                logger.debug("Product's current version:{0} "
                             "is less then first photon version:{1}".format(currentVersionStr,
                                                                            EnvironmentInfo.MINIMUM_ACCEPTABLE_VERSION_STR))
                return False, [{'key': 'inplace_upgrade_version_check.description',
                                'params': [
                                    EnvironmentInfo.get_product_version(humanReadable=True)]}], WARNING_RESULT_TYPE
            if currentVersion >= EnvironmentInfo.INITIAL_PHOTON_VERSION_INT:
                logger.debug("Product's current version:{0} "
                             "is greater or equal to first photon version:{1}".format(currentVersionStr,
                                                                                      EnvironmentInfo.MINIMUM_ACCEPTABLE_VERSION_STR))
                return False, [{'key': 'photon2photon_version_check.description',
                                'params': [
                                    EnvironmentInfo.get_product_version(humanReadable=True)]}], WARNING_RESULT_TYPE
        else:
            logger.error('Could not get products current version.')
            return False, [{'key': 'running_version_check_failed.description'}]

    @staticmethod
    def verify_maximum_directories_count():
        """
        Checks subdirectories count to be under 30k in /tmp
        """
        logger.debug('Entering check minimum directories counts to be under 30K in /tmp.')
        maxDirectoriesCountUnderTmp = 30000
        result = True
        result_description = []

        popen = subprocess.Popen("ls -D /tmp 2>/dev/null | wc -l", shell=True, stdout=subprocess.PIPE,
                                 universal_newlines=True)
        dirCount = popen.stdout.read()
        if int(dirCount) > maxDirectoriesCountUnderTmp:
            logger.error('Subdirectories count under /tmp should be 30K but is {0}'.format(dirCount))
            result = False
            result_description = [{'key': 'verify_maximum_directories_count_failed.description', 'params': [dirCount]}]
            return result, result_description

        return result, result_description

    @staticmethod
    def is_photon():
        return os.path.isfile('/etc/photon-release') and 'photon' in os.uname()[3]

    @staticmethod
    def configs_existence_check():
        """
        Check config files existense
        :return: Boolean with a message
        """
        result = True
        result_description = []
        params = []
        if 'ALIVE_BASE' in os.environ:
            alive_base_path = os.environ['ALIVE_BASE']
        else:
            alive_base_path = '/usr/lib/vmware-vcops'

        # Configuration files
        platform_state_props = '/usr/lib/vmware-vcopssuite/utilities/sliceConfiguration/data/platformState.properties'
        role_state_props = '/usr/lib/vmware-vcopssuite/utilities/sliceConfiguration/data/roleState.properties'
        automation_user_props = os.path.join(alive_base_path, 'user/conf/automationuser.properties')
        persistence_properties = os.path.join(alive_base_path, 'user/conf/persistence/persistence.properties')
        cluster_master_key = os.path.join(alive_base_path, 'user/conf/cluster_master_key.txt')
        maintenance_user_props = os.path.join(alive_base_path, 'user/conf/maintenanceuser.properties')
        migration_user_props = os.path.join(alive_base_path, 'user/conf/migrationuser.properties')
        admin_user_props = '/storage/vcops/user/conf/adminuser.properties'
        cis_props = os.path.join(alive_base_path, 'user/conf/cis.properties')
        inbound = os.path.join(alive_base_path, 'user/plugins/inbound')
        outbound = os.path.join(alive_base_path,
                                'user/plugins/outbound/vrops-generic-rest-plugin/conf/servicenow/config.json')

        # for root and admin password
        shadow = '/etc/shadow'
        # vpostgres
        var_vpostgres = '/var/vmware/vpostgres/'
        opt_vpostgres = '/opt/vmware/vpostgres/'

        hosts = '/etc/hosts'
        host_name_photon = '/etc/hostname'
        host_name_VA = '/etc/HOSTNAME'

        # vami xml
        vami_ovf_info_xml = "/opt/vmware/etc/vami/vami_ovf_info.xml"
        vami_xml = "/opt/vmware/etc/vami/vami.xml"
        ovfEnv_xml = "/opt/vmware/etc/vami/ovfEnv.xml"

        if not os.path.exists(platform_state_props):
            result = False
            params.append(platform_state_props)
        if not os.path.exists(role_state_props):
            result = False
            params.append(role_state_props)
        if not os.path.exists(automation_user_props):
            result = False
            params.append(automation_user_props)
        if not os.path.exists(persistence_properties):
            result = False
            params.append(persistence_properties)
        if not os.path.exists(cluster_master_key):
            result = False
            params.append(cluster_master_key)
        if not os.path.exists(maintenance_user_props):
            result = False
            params.append(maintenance_user_props)
        if not os.path.exists(migration_user_props):
            result = False
            params.append(migration_user_props)
        if not os.path.exists(admin_user_props):
            result = False
            params.append(admin_user_props)
        if not os.path.exists(cis_props):
            result = False
            params.append(cis_props)
        if not os.path.exists(inbound):
            result = False
            params.append(inbound)
        if not os.path.exists(outbound):
            result = False
            params.append(outbound)
        if not os.path.exists(shadow):
            result = False
            params.append(shadow)
        if not os.path.exists(var_vpostgres):
            result = False
            params.append(var_vpostgres)
        if not os.path.exists(opt_vpostgres):
            result = False
            params.append(opt_vpostgres)
        if not os.path.exists(hosts):
            result = False
            params.append(hosts)
        if not os.path.exists(vami_ovf_info_xml):
            result = False
            params.append(vami_ovf_info_xml)
        if not os.path.exists(vami_xml):
            result = False
            params.append(vami_xml)
        if not os.path.exists(ovfEnv_xml):
            result = False
            params.append(ovfEnv_xml)
        if EnvironmentInfo.is_photon():
            if not os.path.exists(host_name_photon):
                result = False
                params.append(host_name_photon)
        else:
            if not os.path.exists(host_name_VA):
                result = False
                params.append(host_name_VA)

        if not result:
            result_description = [{'key': 'check_configs_existence_failed.description', 'params': params}]
        return result, result_description


class Linux(EnvironmentInfo):
    def __init__(self):
        self.logDir = '/var/log/vmware/vcops'

    MAX_LOGIN_FAILURES = 5
    LOCKOUT_TIME = 5 * 60 * 1000
    OS_MAX_LOGIN_FAILURES = 3
    OS_UNLOCK_TIME = 86400
    OS_ROOT_ULOCK_TIME = 300

    def check_httpd_processes_running(self):
        """
        Determines if HTTPD is in a bad state that may break the upgrade.  If it is in a bad state, return a failure
        """
        result = True
        result_description = []
        serviceConfigTestCommand = None
        apache_service_name = None
        logger.info('Running httpd process running check')

        isPhoton = os.path.isfile('/etc/photon-release') and 'photon' in os.uname()[3]
        if isPhoton:
            apache_service_name = 'httpd'
            serviceConfigTestCommand = '/usr/sbin/httpd -t'
        else:
            apache_service_name = 'apache2'
            serviceConfigTestCommand = '/sbin/service ' + apache_service_name + ' configtest'

        if not self.run_config_test_on_service(apache_service_name, serviceConfigTestCommand):
            logger.debug('The apache service name is: %s', apache_service_name)
            result = False
            result_description = [{'key': 'apache_service_name_find_failed.description'}]
            return result, result_description

        logger.debug('Check if the: "%s" service is running', apache_service_name)
        if envInfo.run_script('/sbin/service {0} status'.format(apache_service_name))[0]:
            result_description = [
                {'key': 'check_httpd_processes_running.description', 'params': [apache_service_name.title()]}]
            return result, result_description
        logger.debug('%s is not running, check for processes running', apache_service_name)

        # the service is stopped, determine if there are any httpd processes running.  httpd shows for both service names
        test_result = envInfo.run_script('/bin/ps ax | /bin/grep httpd | /bin/grep -v grep', acceptableReturnCodes=[1])[
            0]
        result = test_result
        result_description = [{'key': 'upgrade_process_failed.description'}]
        return result, result_description

    def run_config_test_on_service(self, service_name, serviceConfigTestCommand):
        if os.geteuid() != 0:
            serviceConfigTestCommand = ''.join(('sudo -n ', serviceConfigTestCommand))

        retval = envInfo.run_script(serviceConfigTestCommand, suppressErrors=True)
        if retval[1] == 0:
            logger.debug('Service: %s config check passed', service_name)
            return True
        else:
            logger.debug('Service: %s config check failed.  Do secondary checks.', service_name)
            stderr = retval[2]
            if stderr.find("no such service") != -1:
                logger.debug('Service: %s - no such service', service_name)
                return False
            if stderr.find("Syntax error") != -1:
                logger.debug('Service: %s - configuration error, but service exists.', service_name)
                return True

            logger.debug('Service: %s - unknown status, assume false.', service_name)
            return False

    def return_free_space(self, directory):
        """
        Returns the free space available in the specified directory
        @param directory directory to return the free space available for
        @return free space available in bytes
        """
        if not os.path.isdir(directory):
            logger.error(
                'Failed to get free space for the directory: "{0}" because it does not exist'.format(directory))
            return None
        stat = os.statvfs(directory)
        freeSpace = stat.f_bavail * stat.f_frsize
        logger.debug('Directory: "{0}" has "{1}" free bytes available'.format(directory, freeSpace))
        return freeSpace

    def get_partitions_for_path(self, path):
        """
        Returns the partitions in which a path resides
        :param path: path to investigate
        :return: path to partition or None in case of error.
        """

        if not path.startswith(os.sep):
            logger.error("Path is not absolute: %s", path)
            return None

        use_shell = False
        df_binary_possibilities = ['/usr/bin/df', '/bin/df']
        for path in df_binary_possibilities:
            if os.path.isfile(path):
                df_command = "{0} -BM -P".format(path)
                break
        else:
            # Give up on absolute path and hope the shell has the command in its path
            logger.warn("Using shell to execute 'df'")
            df_command = "df -BM -P"
            use_shell = True

        logger.debug("Using command '%s'", df_command)

        (success, return_code, df_error, df_output) = self.run_script(shlex.split(df_command), useShell=use_shell)
        if not success:
            logger.error("Unable to execute '%s'\nstdout: %s\nstderr: %s", df_command, df_output, df_error)
            return None

        logger.debug("Output from command '%s':\n%s", df_command, df_output)

        result = None
        for line in df_output.split('\n'):
            if line.startswith('/') or line.startswith('rootfs'):
                components = line.split()
                if len(components) > 5 and components[5].startswith('/'):
                    logger.debug("Found partition: %s", components[5])
                    result = components[5]
                    break

        return result

    def swap_space_check(self):
        """
        Checking if a lot of swap space is being used
        :return:
        """
        result = True

        command = ['free', '-m']
        popen = subprocess.Popen(command, stdout=subprocess.PIPE, universal_newlines=True)
        lines_iterator = iter(popen.stdout.readline, "")
        lastline = ''
        for line in lines_iterator:
            lastline = line

        logger.debug('{0}'.format(lastline))
        lastline = re.sub(' +', ' ', lastline)
        lastline = lastline.split(' ')
        total = lastline[1]
        used = lastline[2]
        logger.debug('total: {0}mb, used: {1}mb'.format(total, used))
        swap_percent = float(used) / float(total)
        if swap_percent > self.SWAP_LIMIT:
            result = False
            # TODO Log top -n 1 here

        result_description = [{'key': 'swap_space_check.description', 'params': [int(swap_percent * 100)]}]

        return result, result_description

    def latency_check_data_nodes(self):
        """
        Checks latency between the nodes
        :return: Boolean with a message
        """
        result = True
        result_description = []

        total = 0
        total_count = 0
        for slice in self.get_slice_info():
            roles = slice.slice_roles
            if EnvironmentInfo.REMOTE_COLLECTOR_ROLE in roles:
                continue

            avg = self.ping_avg_ms(slice.slice_address)
            logger.debug('Average ping time to slice "{0}" is {1}ms'.format(slice.slice_address, avg))
            avg = float(avg)
            total += avg
            total_count += 1

            if avg > self.MAX_LATENCY_DATA_NODE_MS:
                result = False
                result_description = [
                    {'key': 'latency_check_data_nodes_failed.description', 'params': [slice.slice_address, avg]}]
                break
        if result and total_count != 0:
            result_description = [{'key': 'latency_check_data_nodes.description', 'params': [total / total_count]}]

        return result, result_description

    def ping_avg_ms(self, hostname):
        """
        Pings hostname for 5 times, assuming Windows gives this format back
        rtt min/avg/max/mdev = 0.048/0.048/0.048/0.000 ms
        :param hostname: Hostname to ping to
        :return: average response time in millis
        """
        (success, return_code, df_error, out) = self.run_script('/opt/vmware/bin/ovfenv --key [forceIpv6]')
        if out.strip() == 'True':
            if hostname.startswith('[') and hostname.endswith(']'):
                hostname = hostname[1:-1]
            command = ['ping6', '-c', '5', hostname]
        else:
            command = ['ping', '-c', '5', hostname]
        popen = subprocess.Popen(command, stdout=subprocess.PIPE, universal_newlines=True)
        lines_iterator = iter(popen.stdout.readline, "")
        lastline = ''
        for line in lines_iterator:
            lastline = line

        matcher = re.compile("[a-z]+ min/avg/max/[a-z]+ = (\d+.\d+)/(\d+.\d+)/(\d+.\d+)/(\d+.\d+)")
        avg = matcher.search(lastline).groups()[1]
        return avg

    def get_latest_validation_log(self):
        log_file_location = '/var/log/vmware/vcops'
        latest_log = None
        try:
            latest_log = max(glob.iglob(os.path.join(log_file_location, 'vcops_upgrade_validation*.log')),
                             key=os.path.getmtime)
        except Exception as ex:
            logger.error('Exception while getting latest validation log')

        logger.debug('Latest validation log file "{0}"'.format(latest_log))
        return latest_log

    def is_admin_account_lock_from_adminuser(self):
        failed_attempts_key = "failed_attempts"
        time_last_failure_key = "time_last_failure"

        failed_attempts_info = vcops_pack_utilities.getProperties('/storage/vcops/user/conf/adminuser.properties',
                                                                  [failed_attempts_key, time_last_failure_key])
        logger.debug('Failed attempts info : {0}'.format(failed_attempts_info))
        if failed_attempts_key in failed_attempts_info and failed_attempts_info[failed_attempts_key]:
            failed_attempts_value = int(failed_attempts_info[failed_attempts_key])
            if failed_attempts_value > self.MAX_LOGIN_FAILURES:
                current_time_millis = int(round(time.time() * 1000))
                if time_last_failure_key in failed_attempts_info and failed_attempts_info[time_last_failure_key]:
                    time_last_failure_value = int(failed_attempts_info[time_last_failure_key])
                    if time_last_failure_value + self.LOCKOUT_TIME > current_time_millis:
                        return True
        return False

    def root_account_check(self):
        result = True
        result_description = []
        user = "root"

        if self.is_password_empty(user):
            return result, result_description

        if self.is_valid_hash(user):
            result = False
            result_description = [{'key': 'root_password_hash_invalid.failed.description'}]
            return result, result_description

        return result, result_description

    def admin_account_check(self):
        result = True
        new_result_type = None
        result_description = []
        user = "admin"

        if self.is_password_empty(user):
            return result, result_description

        if self.is_account_locked(user):
            result = False
            result_description = [{'key': 'admin_account_locked.failed.description'}]
            return result, result_description
        # OS
        if self.is_account_locked_because_of_failed_attempts(user):
            result = False
            result_description = [{'key': 'admin_account_locked_many_attempts_os.failed.description'}]
            return result, result_description
        # UI
        if self.is_admin_account_lock_from_adminuser():
            result = False
            result_description = [{'key': 'admin_account_locked_many_attempts_ui.failed.description'}]
            return result, result_description
        if self.is_valid_hash(user):
            result = False
            result_description = [{'key': 'admin_password_hash_invalid.failed.description'}]
            return result, result_description

        return result, result_description

    def sda_partition_size_check(self):
        # 21474836480 bytes = 20 GB
        return self._partition_size_check(self.get_partition_size(os.path.join('/', 'dev', 'sda')),
                                          21474836480,
                                          'sda_partition_size_check.failed.description')

    def sdc_partition_size_check(self):
        # 4294967296 bytes = 4 GB
        return self._partition_size_check(self.get_partition_size(os.path.join('/', 'dev', 'sdc')),
                                          4294967296,
                                          'sdc_partition_size_check.failed.description')

    def _partition_size_check(self, currentSize, maxSize, failureMessageKey):
        if not currentSize:
            return False, [{'key': 'unable_to_check.description'}], WARNING_RESULT_TYPE
        if currentSize < maxSize:
            return False, [{'key': failureMessageKey}]
        return True, []

    # TODO make global
    def get_partition_size(self, path):
        cmd = 'lsblk  --output SIZE -n -d -b ' + path
        cmd_result = envInfo.run_script(cmd)

        if not cmd_result[0]:
            logger.error("Unable to execute '%s'\nstdout: %s\nstderr: %s", cmd, cmd_result[3], cmd_result[2])
            return None

        return int(cmd_result[3])

    def inbound_size_check(self):
        plugins_inbound_path = VcOpsPaths.get_plugins_inbound_path()
        inbound_size = self.get_folder_size(plugins_inbound_path)
        inbound_size_human_readable = self.get_folder_size_human_readable(plugins_inbound_path)
        # inbound_max_size = 1288490188 bytes = 1.2 GB
        inbound_max_size = 1288490188

        if not inbound_size:
            return False, [{'key': 'unable_to_check.description'}], WARNING_RESULT_TYPE

        if inbound_size >= inbound_max_size:
            result = False
            logger.info(plugins_inbound_path + " Size is more than acceptable."
                                               " Current size is {0} bites,"
                                               " max acceptable size is {1} bites.".format(inbound_size,
                                                                                           inbound_max_size))
            pathsToClean = [os.path.join(plugins_inbound_path, 'agent_adapter', 'work', 'bundles'),
                            os.path.join(plugins_inbound_path, 'agent_adapter', 'conf', 'plugins', 'agent_plugins',
                                         'bundles')]
            noPathIsCleaned = True
            for path in pathsToClean:
                if os.path.isdir(path):
                    logger.warning('Deleting files and folders inside {0}'.format(path))
                    shutil.rmtree(path)
                    os.mkdir(path)
                    noPathIsCleaned = False
                else:
                    logger.warning('Can not find {0} for deleting its files and folders.'.format(path))

            if noPathIsCleaned:
                return result, [{'key': 'inbound_size_check.failed.description',
                                 'params': [inbound_size_human_readable]}]

            # Check inbound size again after file cleanup
            new_inbound_size = self.get_folder_size(plugins_inbound_path)
            if not new_inbound_size:
                return result, [{'key': 'unable_to_check.description'}], WARNING_RESULT_TYPE

            new_inbound_size_human_readable = self.get_folder_size_human_readable(plugins_inbound_path)
            if new_inbound_size >= inbound_max_size:
                return result, [{'key': 'inbound_size_check.failed.description',
                                 'params': [new_inbound_size_human_readable]}]
            else:
                # size was more than acceptable, but is fixed now
                return result, [{'key': 'inbound_size_check.updated.description',
                                 'params': [new_inbound_size_human_readable]}], WARNING_RESULT_TYPE
        else:
            return True, []

    def vra_7x_check(self):
        try:
            data = self.getCasaManager().get_casa_call('/suite-api/api/adapters?adapterKindKey=VCACAdapter')
        except RestException as e:
            logger.error("Exception During GET REST call /suite-api/api/adapters?adapterKindKey=VCACAdapter")
            logger.error(str(e))
            return False, [{'key': 'unable_to_check.description'}], WARNING_RESULT_TYPE
        except Exception as e:
            logger.error("Unknown error")
            logger.error(str(e))
            return False, [{'key': 'unable_to_check.description'}], WARNING_RESULT_TYPE
        if not data:
            logger.error("No data - VCACAdapte is not installed")
            return False, [{'key': 'unable_to_check.description'}], WARNING_RESULT_TYPE
        else:  # result is not empty
            logger.debug(data)
            for adapterInstancesInfoDto in data["adapterInstancesInfoDto"]:
                if adapterInstancesInfoDto["resourceKey"]["name"] == "vRA76":
                    logger.debug("VCACAdapte is installed")
                    return False, [{'key': 'vra_7x_check_failed.description'}], WARNING_RESULT_TYPE

            logger.debug("VCACAdapte is not installed")
            return True, []

    # TODO make global
    def get_folder_size_human_readable(self, path):
        return self.__get_folder_size(path, True).strip('\n')

    # TODO make global
    def get_folder_size(self, path):
        return int(self.__get_folder_size(path))

    # TODO make global
    def __get_folder_size(self, path, humanReadable=False):
        # du -s /usr/lib/vmware-vcops/user/plugins/inbound | awk '{ print $1}'
        humanReadableStr = 'h ' if humanReadable else ' '
        cmd = 'du -sb' + humanReadableStr + path + ' | awk \'{ print $1}\''
        cmd_result = envInfo.run_script(cmd)

        if not cmd_result[0]:
            logger.error("Unable to execute '%s'\nstdout: %s\nstderr: %s", cmd, cmd_result[3], cmd_result[2])
            return None

        return cmd_result[3]

    def is_account_locked(self, user):

        logger.debug('Checking if account is locked for {0} user'.format(user))
        try:
            enc_pwd = spwd.getspnam(user)[1][0]
            if enc_pwd in "!":
                logger.debug('Account is locked for {0} user'.format(user))
                return True
        except KeyError as e:
            logger.error('User "{0}" not found, value error exception: "{1}"'.format(user, str(e)))
            return False

        logger.debug('Account is not locked for {0} user'.format(user))
        return False

    def is_account_locked_because_of_failed_attempts(self, user):
        command = '/sbin/pam_tally2 --user=' + user
        failed_attempts_data = envInfo.run_script(command)
        if not failed_attempts_data[0]:
            logger.error("Unable to execute '%s'\nstdout: %s\nstderr: %s", command, failed_attempts_data[3],
                         failed_attempts_data[2])
            return False
        failures_count = 0
        if failed_attempts_data[3]:
            failed_attempts_data = failed_attempts_data[3].split('\n')[1].split()
            if len(failed_attempts_data) > 1:
                failures_count = int(failed_attempts_data[1])
            if failures_count >= self.OS_MAX_LOGIN_FAILURES:
                if len(failed_attempts_data) > 3:
                    last_failure_time = datetime.datetime.strptime(
                        failed_attempts_data[2] + ' ' + failed_attempts_data[3], '%m/%d/%y %H:%M:%S')
                    if user == "root":
                        return last_failure_time + datetime.timedelta(
                            seconds=self.OS_ROOT_ULOCK_TIME) >= datetime.datetime.now()
                    return last_failure_time + datetime.timedelta(
                        seconds=self.OS_UNLOCK_TIME) >= datetime.datetime.now()
        return False

    def is_password_empty(self, user):
        logger.debug('Checking if password is empty {0} user'.format(user))

        try:
            enc_pwd = spwd.getspnam(user)[1]
            if not enc_pwd:
                logger.debug('Password is empty '.format(user))
                return True
        except KeyError as e:
            logger.error('User "{0}" not found, value error exception: "{1}"'.format(user, str(e)))
            return False

    def is_valid_hash(self, user):
        logger.debug(
            'Checking if password hashing algorithm satisfied in-place upgrade requirements for {0} user'.format(user))

        try:
            enc_pwd = spwd.getspnam(user)[1][1]
            if enc_pwd in "2":
                logger.debug('Hashing algorithm is not valid for {0} user'.format(user))
                return True
        except KeyError as e:
            logger.error('User "{0}" not found, value error exception: "{1}"'.format(user, str(e)))
            return False

        logger.debug('Hashing algorithm is valid for {0} user'.format(user))
        return False

    ## Checking ESXi version Bug 2291438
    def esxi_version_check(self):
        is_SAAS = os.environ.get('IS_SAAS')
        if not is_SAAS or 'false' in is_SAAS:
            esx_major_version_out = envInfo.run_script(
                'vmware-toolbox-cmd stat raw text session | {}'.format('sed -n -e "s/.*ESX\(.*\)build-.*/\\1/p"'))
            if not esx_major_version_out[2]:
                esx_major_version = int(esx_major_version_out[3].strip().replace('.', ''))
                if esx_major_version < 600:
                    return False, [{'key': 'esxi_version_check_failed.description'}]
                elif esx_major_version == 650:
                    build_number_out = envInfo.run_script(
                        'vmware-toolbox-cmd stat raw text session | {}'.format('sed -n -e "s/^.*build-//p"'))
                    if not build_number_out[2]:
                        build_number_6_5_GA = 4564106
                        build_number_6_5_U1 = 5969303
                        if build_number_6_5_GA <= int(build_number_out[3]) < build_number_6_5_U1:
                            return False, [{'key': 'esxi_version_check_failed.description'}]
            else:
                return False, [{'key': 'unable_to_check.description'}], WARNING_RESULT_TYPE
        return True, []


class VcOpsPaths(object):
    @staticmethod
    def find_vcops_base_path():
        """Finds the VCOPS_BASE environment variable.
        @return: the VCOPS_BASE path or an exception if it cannot be found.
        """
        vcops_base_path = os.environ.get('VCOPS_BASE')
        if not vcops_base_path:
            msg = (
                'Environment variables "VCOPS_BASE" is not set.'
                'Please load the vC Ops environment and try again.'
            )
            sys.stderr.write('\n' + msg)
            raise KeyError(msg)

        if not os.path.isdir(vcops_base_path):
            msg = 'VCOPS_BASE path: "' + vcops_base_path + '" is not a directory'
            sys.stderr.write('\n' + msg)
            raise IOError(msg)

        return vcops_base_path

    @staticmethod
    def find_storage_db_path():
        """Finds the STORAGE_DB_VCOPS environment variable.
        @return: the STORAGE_DB_VCOPS path or an exception if it cannot be found.
        """
        storage_db_path = os.environ.get('STORAGE_DB_VCOPS')
        if not storage_db_path:
            msg = (
                'Environment variable "STORAGE_DB_VCOPS" is not set.'
                'Please load the vC Ops environment and try again.'
            )
            sys.stderr.write('\n' + msg)
            raise KeyError(msg)

        if not os.path.isdir(storage_db_path):
            msg = 'STORAGE_DB_VCOPS path: "' + storage_db_path + '" is not a directory'
            sys.stderr.write('\n' + msg)
            raise IOError(msg)

        return storage_db_path

    @staticmethod
    def find_storage_core_path():
        """Finds the STORAGE_CORE_VCOPS environment variable.
        @return: the STORAGE_CORE_VCOPS path or an exception if it cannot be found.
        """
        storage_core_path = os.environ.get('STORAGE_CORE_VCOPS')
        if not storage_core_path:
            msg = (
                'Environment variable "STORAGE_CORE_VCOPS" is not set.'
                'Please load the vC Ops environment and try again.'
            )
            sys.stderr.write('\n' + msg)
            raise KeyError(msg)

        if not os.path.isdir(storage_core_path):
            msg = 'STORAGE_CORE_VCOPS path: "' + storage_core_path + '" is not a directory'
            sys.stderr.write('\n' + msg)
            raise IOError(msg)

        return storage_core_path

    @staticmethod
    def find_storage_log_path():
        """Finds the STORAGE_LOG_VCOPS environment variable.
        @return: the STORAGE_LOG_VCOPS path or an exception if it cannot be found.
        """
        storage_log_path = os.environ.get('STORAGE_LOG_VCOPS')
        if not storage_log_path:
            msg = (
                'Environment variable "STORAGE_LOG_VCOPS" is not set.'
                'Please load the vC Ops environment and try again.'
            )
            sys.stderr.write('\n' + msg)
            raise KeyError(msg)

        if not os.path.isdir(storage_log_path):
            msg = 'STORAGE_LOG_VCOPS path: "' + storage_log_path + '" is not a directory'
            sys.stderr.write('\n' + msg)
            raise IOError(msg)

        return storage_log_path

    @staticmethod
    def find_storage_path():
        """Finds the STORAGE environment variable.
        @return: the STORAGE path or an exception if it cannot be found.
        """
        storage_path = os.environ.get('STORAGE')
        if not storage_path:
            msg = (
                'Environment variable "STORAGE" is not set.'
                'Please load the vC Ops environment and try again.'
            )
            sys.stderr.write('\n' + msg)
            raise KeyError(msg)

        if not os.path.isdir(storage_path):
            msg = 'STORAGE path: "' + storage_path + '" is not a directory'
            sys.stderr.write('\n' + msg)
            raise IOError(msg)

        return storage_path

    @staticmethod
    def get_plugins_inbound_path():
        return os.path.join(VcOpsPaths.find_vcops_base_path(), 'user', 'plugins', 'inbound')


def setup_logger():
    # setup the logger
    global logger
    logger = logging.getLogger('b2b')
    logger.setLevel(logging.DEBUG)
    envInfo = EnvironmentInfo.newInstance()
    # set log dir
    if not os.path.isdir(envInfo.logDir):
        try:
            os.makedirs(envInfo.logDir)
        except EnvironmentError:
            logger.exception('Failed to make log dir "%s"', envInfo.logDir)
            sys.exit(1)
    # create the file logger
    filename = os.path.join(envInfo.logDir, 'vcops_upgrade_validation_{0}.log'.format(time.strftime("%Y%m%d-%H%M%S")))
    mode = 'a'
    maxBytes = 1048576
    backupCount = 5
    fh = logging.handlers.RotatingFileHandler(filename, mode, maxBytes, backupCount)
    fh.setLevel(logging.DEBUG)
    fileFormatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(module)s - %(funcName)s - %(message)s')
    fh.setFormatter(fileFormatter)
    logger.addHandler(fh)
    # create the console logger
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    consoleFormatter = logging.Formatter('%(levelname)s - %(message)s')
    ch.setFormatter(consoleFormatter)
    logger.addHandler(ch)


class CaSARestException(Exception):
    response_code = None
    def __init__(self, response_code, errorMessage):
        Exception.__init__(self, errorMessage)
        self.response_code = response_code


class CasaManager(object):
    _connection = None
    headers = {'Content-Type': 'application/json', 'Accept': 'application/json', 'X-vRealizeOps-API-use-unsupported': 'true'}
    envInfo = EnvironmentInfo.newInstance()

    def __init__(self):
        self.token_auth_header = None

    def get_connection(self, hostname=None):
        if not hostname:
            hostname = 'localhost'
        if not self._connection:
            if not hasattr(ssl, 'SSLContext'):
                self._connection = vcops_pak_six.VcopsHTTPSConnection(hostname, timeout=60)
            else:
                self.envInfo = EnvironmentInfo.newInstance()
                cafile = os.path.join(VcOpsPaths.find_storage_path(), 'vcops', 'user', 'conf', 'ssl', 'web_chain.pem')
                with open(cafile, 'r') as f:
                    cadata = f.read()
                if not cadata:
                    cafile = os.path.join(VcOpsPaths.find_storage_path(), 'vcops', 'user', 'conf', 'ssl',
                                          'web_cert.pem')
                sslcontext = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH, cafile=cafile)
                # not ideal - but the alternative is to check vs. thumbprints gathered from the same place and the trusted cert list
                sslcontext.check_hostname = False
                self._connection = vcops_pak_six.VcopsHTTPSConnection(hostname, context=sslcontext)

        return self._connection

    def return_maintenance_user_credentials(self):
        """
        Read the maintenanceuser.properties file and generate a corresponding
        HTTP Authentication header value from it. If the file doesn't exist,
        """
        maintenance_user_file = os.path.join(VcOpsPaths.find_vcops_base_path(), 'user', 'conf', 'maintenanceuser.properties')
        if os.path.isfile(maintenance_user_file):
            username = self.envInfo.load_dictionary_key_value(maintenance_user_file, 'username')
            password = self.envInfo.load_dictionary_key_value(maintenance_user_file, 'password')
            encrypted = self.envInfo.load_dictionary_key_value(maintenance_user_file, 'encrypted')

            if not encrypted or encrypted.lower() == 'true':
                password = Crypt().decrypt(password)

            return username, password
        else:
            logger.error('Maintenance User credentials not found', exitCode=1)

    def make_auth_header(self, user_name, password):
        if vcops_pak_six.VCOPS_PY3:
            up = user_name + ':' + password
            up = up.encode('utf-8')
            # pylint:disable=no-member
            auth_token = base64.b64encode(up)
            # pylint:enable=no-member
            auth_token = auth_token.decode('utf-8')
            auth_header = 'Basic {0}'.format(auth_token)
        else:
            auth_token = base64.encodestring('{0}:{1}'.format(user_name, password))[:-1]  # strip appended newline
            auth_header = "Basic {0}".format(auth_token)
        return auth_header

    class Server(Enum):
        CASA = 1
        SUITE_API = 2

    class CallType(Enum):
        GET = 1
        POST = 2

    def get_casa_call(self, url, logger=None, log_response=True):
        global token_pair
        retry = 0
        retries = 10
        polling_delay = 10  # seconds
        while retry < retries:
            response_code, response, token_pair = vcops_pack_http_utilities.getUrl(url,
                                                    token_pair=token_pair,
                                                    credential_retriever=self.return_maintenance_user_credentials)
            if response_code != vcops_pack_http_utilities.HTTP_OK:
                logger.info('Found status code: {0}, ignore it and try again after 10 seconds'.format(response_code))
                time.sleep(polling_delay)
                retry += 1
            else:
                break
        # pass the response to the handler even in the case when retries didn't help
        return CasaManager.handle_http_response(url, response_code, response, logger, log_response=log_response)

    def post_casa_call(self, url, request_data, log_response=True):
        global token_pair
        response_code, response, token_pair = vcops_pack_http_utilities.postUrl(url, values=request_data,
                                                    additional_headers=self.headers,
                                                    token_pair=token_pair,
                                                    credential_retriever=self.return_maintenance_user_credentials)
        return CasaManager.handle_http_response(url, response_code, response, logger, log_response=log_response)

    @classmethod
    def handle_http_response(cls, url, response_code, response, logger=None, log_response=True):
        if response_code != vcops_pack_http_utilities.HTTP_OK:
            error_message = 'Failed to call CaSA url {0}. CaSA returned status code {1}, response: {2}'. \
                format(url, response_code, response)
            logger.error(error_message)
            raise CaSARestException(response_code, error_message)

        data = json.loads(response)
        if data and log_response:
            logger.debug('Response from url {0}'.format(url))
            logger.debug(json.dumps(data, indent=4, sort_keys=True))
        return data

    def get_suite_api_call(self, url, hostname=None, log_response=True):
        return self.suite_api_call(self.Server.SUITE_API, self.CallType.GET, url, hostname=hostname, log_response=log_response)

    def post_suite_api_call(self, url, request_data, hostname=None, log_response=True):
        return self.suite_api_call(self.Server.SUITE_API, self.CallType.POST, url, request_data=request_data, hostname=hostname, log_response=log_response)

    def suite_api_call(self, server, call_type, url, request_data=None, hostname=None, log_response=True):
        self.set_auth_header(server, url)

        if hostname:
            hostname = hostname.lower()
            connection = self.get_connection(hostname)
        else:
            connection = self.get_connection()
        if request_data:
            connection.request(call_type.name, url, body=request_data, headers=self.headers)
        else:
            connection.request(call_type.name, url, headers=self.headers)
        connection.set_debuglevel(2)
        response = connection.getresponse()

        if response.status != 200:
            errorMsg = 'Failed to call {0} url {1}. {0} returned status code {2}, response: {3}'.\
                format(server.name, url, response.status, response.read())
            logger.error(errorMsg)
            raise RestException(response, errorMsg)

        data = json.loads(response.read().decode('utf-8'))
        if data and log_response:
            logger.debug('Response from url {0} to host {1}'.format(url, hostname))
            logger.debug(json.dumps(data, indent=4, sort_keys=True))
        return data

    def set_auth_header(self, server, url):
        if url == "/suite-api/api/auth/token/acquire":
            if 'Authorization' in self.headers:
                del self.headers['Authorization']
        else:
            if not self.token_auth_header:
                authentication_token = self.acquire_authentication_token()
                self.token_auth_header = 'vRealizeOpsToken {token}'.format(token=authentication_token)

            self.headers['Authorization'] = self.token_auth_header

    def acquire_authentication_token(self, verify=False):
        suite_api_token_acquire_url = "/suite-api/api/auth/token/acquire"

        username, password = self.return_maintenance_user_credentials()
        body = {
           "username": username,
           "password": password,
        }

        request_data = json.dumps(body)
        data = CasaManager().post_suite_api_call(suite_api_token_acquire_url, request_data, log_response=False)
        if not data:
            raise NoDataException(suite_api_token_acquire_url)

        return data["token"]

    def get_latency_for_call(self, url, hostname, authorization_required):
        current_time_millis = lambda: int(round(time.time() * 1000))
        url = vcops_pack_http_utilities.create_casa_url(url, hostname)

        global token_pair
        start_time = current_time_millis()
        response_code, response, token_pair = vcops_pack_http_utilities.getUrl(url,
                                                                   authorization_required=authorization_required,
                                                                   additional_headers=self.headers,
                                                                   token_pair=token_pair)
        if response_code != vcops_pack_http_utilities.HTTP_OK:
            raise RuntimeError('Failed to ping. Status code is {0}'.format(response_code))
        end_time = current_time_millis()

        return end_time - start_time

    def ping_node_to_find_admin(self, logger=None):
        logger.info('Identifying if the node is MASTER with security ping')
        result = False

        url = '/cluster/security/ping'
        casa_ping_url = vcops_pack_http_utilities.create_casa_url(url)
        response = CasaManager().get_casa_call(casa_ping_url, logger=logger)

        if not response:
           program_exit('Failed couldn\'t identify node type', 1)

        if 'is_admin_slice' in response:
           if response['is_admin_slice']:
               logger.info('Current node is MASTER node')
               result = True
        else:
           program_exit('Failed couldn\'t identify node type', 1)
        return result


class Check(object):
    def __init__(self, check_name, check_type, start_time=None, stop_time=None, cluster_online=None,
                 result=None, result_description=None, friendly_name=None, run_on_master=None,
                 readable_start_time=None, readable_stop_time=None, recommendation=None, skip_check=None):
        self.check_name = check_name
        self.start_time = start_time
        self.stop_time = stop_time
        self.check_type = check_type
        self.result = result
        self.result_description = result_description
        self.friendly_name = friendly_name
        self.cluster_online = cluster_online
        self.run_on_master = run_on_master
        self.readable_start_time = readable_start_time
        self.readable_stop_time = readable_stop_time
        self.recommendation = recommendation
        self.skip_check = skip_check

    @staticmethod
    def return_times():
        current_time = time.time()
        readable_current_time = datetime.datetime.fromtimestamp(current_time).strftime('%Y-%m-%d %H:%M:%S')
        return int(current_time * 1000), readable_current_time

    def set_start_time(self):
        current_time, readable_current_time = Check.return_times()
        # logger.debug('Set start time to: %s (%s) for check: %s', current_time, readable_current_time, self.check_name)
        self.start_time = current_time
        self.readable_start_time = readable_current_time

    def set_stop_time(self):
        current_time, readable_current_time = Check.return_times()
        # logger.debug('Set STOP time to: %s (%s) for check: %s', current_time, readable_current_time, self.check_name)
        self.stop_time = current_time
        self.readable_stop_time = readable_current_time

    def set_result(self, result):
        # logger.debug('Setting result to: %s for check: %s', result, self.check_name)
        self.result = result

    def set_result_description(self, result_description):
        # logger.debug('Setting result description to: %s for check: %s', result_description, self.check_name)
        self.result_description = result_description

    def update_result_type(self, new_result_type):
        self.check_type = new_result_type

    def return_json(self):
        return self.__dict__


class StateDoc(object):

    def __init__(self):
        self.state_file_path = None
        self.state_doc = None
        self.check_list = []
        self.template_state_file = StateDoc.get_template_state_file_path()
        self.state_file_path = StateDoc.construct_state_file_path()
        if not os.path.isfile(self.state_file_path):
            self.deploy_state_template()
        self.set_state_doc()
        self.load_checks_list()

    @staticmethod
    def get_template_state_file_path():
        template_state_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'preUpgradeNewChecks.json')
        if os.path.isfile(template_state_file):
            return template_state_file
        else:
            logger.error('state file template "{0}" not found'.format(template_state_file))
            return None

    @staticmethod
    def get_mp_file_path():
        mp_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'ManagementPaksList.json')
        if os.path.isfile(mp_file):
            return mp_file
        else:
            logger.error('Management Pak file template "{0}" not found'.format(mp_file))
            return None

    @staticmethod
    def construct_state_file_path():
        pak_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        pak_id = os.path.basename(pak_root)
        state_file_path = os.path.join(pak_root, '{0}_new_validate.json'.format(pak_id))
        # logger.debug('Constructed state file path: "{0}"'.format(state_file_path))
        return state_file_path

    @staticmethod
    def construct_state_file_path_old():
        pak_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        pak_id = os.path.basename(pak_root)
        state_file_path = os.path.join(pak_root, '{0}_validate.json'.format(pak_id))
        # logger.debug('Constructed state file path: "{0}"'.format(state_file_path))
        return state_file_path

    def deploy_state_template(self):
        logger.debug('Entering deploy the state file template: "%s" to: "%s"', StateDoc.get_template_state_file_path(),
                     self.state_file_path)
        if os.path.isfile(StateDoc.get_template_state_file_path()):
            try:
                shutil.copyfile(StateDoc.get_template_state_file_path(), self.state_file_path)
            except EnvironmentError:
                error_message = 'Failed to copy template state file: "{0}" to: "{1}"'.format(
                    StateDoc.get_template_state_file_path(), self.state_file_path)
                logger.exception(error_message)
                raise error_message

    def from_json_file(self):
        self.state_doc = None
        logger.debug('state file: "%s"', self.state_file_path)
        if os.path.isfile(self.state_file_path):
            with open(self.state_file_path) as state_file:
                self.state_doc = json.load(state_file)

    def setup_initial_state_doc(self):
        self.state_doc = json.loads('{"checks": []}')
        return self.state_doc

    def set_state_doc(self):
        if not self.state_doc:
            self.from_json_file()
        if not self.state_doc:
            self.setup_initial_state_doc()
        return self.state_doc

    def load_checks_list(self):
        self.set_state_doc()
        for check in self.state_doc['checks']:
            self.check_list.append(Check(**check))

    def update_start_time(self, check_to_update):
        for check in self.check_list:
            if check.check_name == check_to_update:
                check.set_start_time()

    def update_stop_time(self, check_to_update):
        for check in self.check_list:
            if check.check_name == check_to_update:
                check.set_stop_time()

    def update_result(self, check_to_update, result):
        for check in self.check_list:
            if check.check_name == check_to_update:
                check.set_result(result)

    def update_result_description(self, check_to_update, result_description):
        for check in self.check_list:
            if check.check_name == check_to_update:
                check.set_result_description(result_description)

    def update_result_type(self, check_to_update, new_result_type):
        if new_result_type:
            for check in self.check_list:
                if check.check_name == check_to_update:
                    check.update_result_type(new_result_type)

    def return_state_json(self):
        self.state_doc['checks'] = []
        for check in self.check_list:
            json = check.return_json()
            self.state_doc['checks'].append(json)

    def write_state_doc_actions_update(self):
        self.return_state_json()
        self.write_state_file()

    def write_state_file(self):
        try:
            with open(self.state_file_path, 'w') as working_state_file:
                json.dump(self.state_doc, working_state_file, indent=4, sort_keys=True)
        except EnvironmentError:
            logger.exception('Failed to write to the state file: "%s"', self.state_file_path)


class SliceInfo():
    slice_address = None
    slice_uuid = None
    slice_roles = None

    def __init__(self, slice_address, slice_uuid, slice_roles):
        self.slice_address = slice_address
        self.slice_uuid = slice_uuid
        self.slice_roles = slice_roles


class NoDataException(Exception):
    def __init__(self, casa_url):
        Exception.__init__(self, 'No data returned from CaSA call "{0}"'.format(casa_url))


class RestException(Exception):
    respone = None

    def __init__(self, response, errorMessage):
        Exception.__init__(self, errorMessage)
        self.respone = response


class PakValidator():

    @staticmethod
    def copy_state_file_to_apuat_log_dir(envInfo, state_file, apuat_report_dir):
        logger.debug("copy_state_file_to_apuat_log_dir")
        logger.debug("state_file {0}".format(state_file))
        logger.debug("apuat_report_dir {0}".format(apuat_report_dir))
        if not os.path.exists(state_file):
            logger.error("state_file does not exist in: {0}".format(state_file))
            return
        if not os.path.exists(apuat_report_dir):
            logger.error("apuat_report_dir does not exist in: {0}".format(apuat_report_dir))
            return
        try:
            resultText = "var CHECKS ={0};"
            with open(state_file, "r") as file:
                resultText = resultText.format(file.read())

            state_file_js = "validate_checks.js"
            with open(state_file_js, "w") as file:
                file.write(resultText)

            envInfo.run_script('cp {0} {1} '.format(state_file_js, apuat_report_dir))
        except Exception as e:
            logger.error("error error when moving state_file to apuat_report_dir")
            logger.error(str(e))
        logger.debug("state_file copied to apuat_log_dir")

    @staticmethod
    def main():
        try:
            global envInfo
            envInfo = EnvironmentInfo.newInstance()
            setup_logger()
        except Exception as setup_logs_exception:
            print('Failed to setup logs: {0}'.format(setup_logs_exception))
            sys.exit(1)

        try:
            pak_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            # Copying bypass_prechecks.json from extracted folder to pakrepolocal folder
            # Ideally this should be done in PakManager, but since existing pakmanager is used
            # to stage the pak file, temporarily copying from here till next release
            envInfo.copy_bypass_checks(pak_root)

            # Reading the checks from json file
            data = envInfo.read_from_json_file(StateDoc.get_template_state_file_path())
            if not data:
                logger.error('Couldn\'t get checks from "{0}"'.format(StateDoc.get_template_state_file_path()))
                sys.exit(1)

            # Update pak results file with validate status file path
            envInfo.update_pak_results_file(state_file=True)
            checks = data['checks']
            exit_code = 0

            cluster_online = False
            admin_role_enabled = False
            certificate_expired = False
            try:
                cluster_online = envInfo.is_cluster_online()
            except ssl.SSLError as ex:
                certificate_expired = True
                logger.warning('Certificate is expired many checks will be skippedd.')
                logger.warning(str(ex))
            except Exception as ex:
                logger.info('Checks that require Cluster to be Online will be skipped.')
                logger.exception('Exception while trying to know if the cluster is online. {0}'.format(ex))

            try:
                admin_role_enabled = envInfo.admin_role_enabled()
            except Exception as ex:
                logger.info('Checks that has to be run by ADMIN will be skipped.')
                logger.exception('Exception while trying get slice roles. {0}'.format(ex))

            failed_checks = []
            for check in checks:
                check_name = check['check_name']
                new_result_type = None
                logger.info('***' + check_name + '*** --- ' + check['friendly_name'])

                envInfo.update_state_file(check_name, set_start_time=True)
                if check['cluster_online'] == 'true' and not cluster_online:
                    if certificate_expired:
                        result = False
                        new_result_type = WARNING_RESULT_TYPE
                        result_description = [
                            {'key': 'skipping_check_expired certificate.description', 'params': []}]
                    else:
                        result = True
                        result_description = [
                            {'key': 'skipping_check_cluster_is_not_online.description', 'params': [check_name]}]
                elif check['run_on_master'] == 'true' and not admin_role_enabled:
                    if certificate_expired:
                        result = False
                        new_result_type = WARNING_RESULT_TYPE
                        result_description = [
                            {'key': 'skipping_check_expired certificate.description', 'params': []}]
                    else:
                        result = True
                        result_description = [
                            {'key': 'skipping_check_not_run_on_master.description', 'params': [check_name]}]
                elif check['skip_check'] == "true":
                    result = True
                    result_description = [{'key': 'skipping_check_not_run_on_master.description',
                                           'params': [check_name, StateDoc.get_template_state_file_path()]}]
                else:
                    try:
                        resultTuple = getattr(envInfo, check_name)()
                        if len(resultTuple) >= 2:
                            result = resultTuple[0]
                            result_description = resultTuple[1]
                        if len(resultTuple) == 3:
                            new_result_type = resultTuple[2]
                    except ssl.SSLError as ex:
                        result = False
                        new_result_type = WARNING_RESULT_TYPE
                        result_description = [
                            {'key': 'skipping_check_expired certificate.description', 'params': []}]
                    except Exception as general_exception:
                        result = False
                        new_result_type = WARNING_RESULT_TYPE
                        result_description = [{'key': 'unable_to_check.description', 'params': []}]
                        logger.exception('{0} failed, {1}'.format(check_name, general_exception))
                if result_description:
                    if result:
                        logger.info(result_description)
                    else:
                        logger.error(result_description)
                envInfo.update_state_file(check['check_name'], set_stop_time=True,
                                          result=result, result_description=result_description,
                                          new_result_type=new_result_type)

                if not result:
                    failed_checks.append(check_name)
                    if check_name == 'node_online_check':
                        logger.debug('Checking if the check type changed')
                        pre_check = envInfo.get_check(check_name)
                        if pre_check:
                            check['check_type'] = pre_check['check_type']
                    if check['check_type'] == 'ERROR' and not envInfo.bypass_preupgrade_checks(pak_root):
                        logger.error('{0} failed, update the exit code'.format(check_name))
                        if exit_code == 0:
                            validation_log = envInfo.get_latest_validation_log()
                            resource_directory = os.path.join(pak_root, 'resources')
                            if not validation_log:
                                validation_log = ''
                            # envInfo.update_pak_results_file(exit_code=str(3), failed_check='{0}'.format(check_name),\
                            #                                 message='validate.script.pre_upgrade_check_failure_message',\
                            #                                 args = [check_name, validation_log],\
                            #                                 directory=resource_directory)
                        exit_code = 0
                    else:
                        # Don't stop the upgrade for WARN cases
                        logger.error('{0} failed'.format(check_name))

            if not failed_checks:
                logger.info('All pre-upgrade checks got passed')
            else:
                logger.info('Pre-upgrade checks "{0}" got failed, please check at "{1}"'
                            .format(failed_checks, StateDoc.construct_state_file_path()))
            if len(sys.argv) > 1:
                PakValidator.copy_state_file_to_apuat_log_dir(envInfo, StateDoc.construct_state_file_path(),
                                                              sys.argv[1])
            sys.exit(exit_code)
        except Exception as general_exception:
            logger.error('Failed running upgrade: {0}'.format(general_exception))
            sys.exit(1)
        finally:
            vcops_pack_http_utilities.logout("localhost", token_pair)


if __name__ == "__main__":
    PakValidator.main()
