import os
import sys
from base64 import b64decode

try:
    from Crypto.Cipher import AES
except ImportError:
    try:
        from cryptography.hazmat.primitives.ciphers.algorithms import AES
        from cryptography.hazmat.primitives.ciphers import Cipher, modes
        from cryptography.hazmat.backends import default_backend as backend
    except ImportError:
        print(__name__ + ': Could not import Crypto.Cipher.AES, probably a test environment')


# This class mimics java com.vmware.vcops.security.Crypt class and python
# .../vsutilities/vmware-vcopssuite/utilities/lib/vcops_crypt.py
# IMPORTANT! In case of any functional change all classes should be changed simultaneously.
class Crypt:
    def __init__(self, master_key_path=None):
        if master_key_path is None:
            self.keyFilePath = os.path.join(Crypt.find_vcops_base_path(), 'user', 'conf', 'cluster_master_key.txt')
        else:
            self.keyFilePath = master_key_path
            
        try:
            self.backend = backend()
        except:
            pass

    def decrypt(self, text):
        if not text:
            print('WARNING: Message is empty, nothing to decrypt.')
            return None
        text = text.strip()

        if ':' in text:
            return self.decrypt_impl_v2(text)
        else:
            if sys.version_info[0] == 3:
                # python 3 implementation
                return self.decrypt_impl_v1(text)
            else:
                # python 2 implementation
                return self.decrypt_impl_old(text)


    def decrypt_impl_v2(self, text):
        parts = text.split(':')
        if len(parts) != 3:
            print('ERROR: Invalid message format, unable to decrypt.')
            return None

        crypt_settings = None
        with open(self.keyFilePath, 'r') as keyFile:
            for line in keyFile:
                cs = self.get_crypt_settings(line)
                if cs is not None and cs['version'].lower() == parts[0].lower():
                    crypt_settings = cs
                    break
        if crypt_settings is None:
            # message is encrypted with unknown key
            print('ERROR: Message is encrypted with unknown key.')
            return None

        iv = b64decode(parts[1])
        ct = b64decode(parts[2])
        try:
            dcipher = AES.new(crypt_settings['key'], AES.MODE_CBC, iv)
            decrypted_data = dcipher.decrypt(ct)
        except:
            dcipher = Cipher(AES(crypt_settings['key']), modes.CBC(iv), backend=self.backend)
            decryptor = dcipher.decryptor()
            decrypted_data = decryptor.update(ct) + decryptor.finalize()

        return str(self.unpad(decrypted_data), 'UTF-8')

    def decrypt_impl_v1(self, text):
        if text[-1] == '@':
            text = text[0:-1]

        crypt_settings = None
        with open(self.keyFilePath, 'r') as keyFile:
            # V1 key should be the first line of the key file
            line = keyFile.readline()
            crypt_settings = self.get_crypt_settings(line)
            if crypt_settings is None or crypt_settings['version'].upper() != 'V1':
                print('ERROR: V1 key is not found in key file {0}'.format(self.keyFilePath))
                return None

        key = crypt_settings['key']
        iv = key
        ct = b64decode(text)
        try:
            dcipher = AES.new(key, AES.MODE_CBC, iv)
            decrypted_data = dcipher.decrypt(ct)
        except:
            dcipher = Cipher(AES(crypt_settings['key']), modes.CBC(iv), backend=self.backend)
            decryptor = dcipher.decryptor()
            decrypted_data = decryptor.update(ct) + decryptor.finalize()

        return str(self.unpad(decrypted_data), 'UTF-8')


    # Python 2.* specific implementation
    def decrypt_impl_old(self, text):
        print("lolik")
        if not self.keyFilePath:
            print('Cluster master key file not found.')

        master_key = None
        try:
            with open(self.keyFilePath, 'r') as mkf:
                master_key = mkf.read()
        except EnvironmentError as e:
            print('Cluster master key file can\'t be read: {0}'.format(str(e)))
            return None

        # The encrypted password is assumed to be wrapped in Base-64 encoding, possibly
        # with an extra trailing '@'. (See com.vmware.vcops.common.security.Crypt ).
        text = text.strip()
        if text[-1] == '@':
            text = text[0:-1]

        try:
            passwordbytes = b64decode(text)
            cipher = AES.new(master_key, AES.MODE_CBC, master_key[:16])
            text = cipher.decrypt(passwordbytes)
            return text[0:-ord(text[-1])]
        except TypeError as e:
            print('Failed to decode password, type error exception: {0}'.format(str(e)))
            return None
        except ValueError as e:
            print('Failed to decode password, value error exception: {0}'.format(str(e)))
            return None


    def get_crypt_settings(self, line):
        line = line.strip()
        if not line:
            return None
        ret = dict()
        parts = line.split(' ')
        if len(parts) == 1:
            # old key format
            key = parts[0].strip()
            ret['version'] = 'V1'
            ret['key'] = bytes(key, 'UTF-8')
        else:
            ret['version'] = parts[0]
            for s in parts[1:]:
                kvPair = s.split('=', 1)
                kvPair[0] = kvPair[0].lower()
                if kvPair[0] == 'key':
                    ret['key'] = b64decode(bytes(kvPair[1], 'UTF-8'))

        return ret


    def pad(self, data):
        length = AES.block_size - (len(data) % AES.block_size)
        data += bytes([length]) * length
        return data


    def unpad(self, data):
        return data[0:-(data[-1])]


    @classmethod
    def find_vcops_base_path(cls):
        """Finds the VCOPS_BASE environment variable.
        @return: the VCOPS_BASE path or an exception if it cannot be found.
        """

        try:
            vcopsBaseDir = os.environ["ALIVE_BASE"]
        except KeyError:
            vcopsBaseDir = os.path.join(os.sep, "usr", "lib", "vmware-vcops")

        return vcopsBaseDir
