# Copyright 2017 VMware, Inc.  All rights reserved. -- VMware Confidential

__author__ = "VMware, Inc."

import base64
import os

import ssl
import socket
import hashlib

import logging
import json
import sys

try:
    vcops_base_dir = os.environ["ALIVE_BASE"]
except KeyError:
    vcops_base_dir = os.path.join(os.sep, "usr", "lib", "vmware-vcops")

utilitiesDir = os.path.join(vcops_base_dir, "..", "vmware-vcopssuite", "utilities")
sys.path.append(os.path.join(utilitiesDir, 'lib'))
import vc_ops_utilities
import vcops_pack_properties

import vcops_pak_six

class HttpRequestErrors(object):
    """Error codes in case of http exception"""
    URLLIB2_URLERROR = 9000
    SOCKET_TIMEOUT = 9001
    SSL_SSLERROR = 9002
    HTTPLIB_HTTPEXCEPTION = 9003
    SERVICE_UNAVAILABLE = 503


REFRESH_TOKEN = "refreshToken"
ACCESS_TOKEN = "accessToken"
DELTA_PERIOD = 1000  # milliseconds

HTTP_OK = 200
HTTP_ACCEPTED = 202

UNAUTHORIZED = 401
NOT_FOUND = 404
PRECONDITION_FAILED = 412

Authorization = "Authorization"


# get python to check a thumbprint instead of certificate
# check happens in httplib.HTTPSConnection::connect -- override this in ThumbprintHTTPSConnection
# The HTTPSConnection is instantiated in urllib2.HTTPSHandler::https_open -- override this in ThumbprintHTTPSHandler
# to change the call to urllib2.do_open
class ThumbprintHTTPSConnection(vcops_pak_six.VcopsHTTPSConnection):
    """
    Override httplib.HTTPSConnection in order to add thumbprint validation instead of truststore based
    validation of the certificate
    """

    _expected_thumbprint = None

    def __init__(self, *args, **kwargs):
        # remove thumbprint before initializing superclass
        if 'thumbprint' in kwargs:
            kwargs = dict(kwargs)
            self._expected_thumbprint = kwargs['thumbprint']
            del kwargs['thumbprint']

        vcops_pak_six.VcopsHTTPSConnection.__init__(self, *args, **kwargs)

    def _validate_thumbprint(self):
        """
        make sure the leaf certificate hash matches the one we've been given
        """
        # pylint:disable=no-member
        derCert = self.sock.getpeercert(True)
        # pylint:enable=no-member
        sha1 = hashlib.sha1()
        sha1.update(derCert)
        sha1Digest = sha1.hexdigest().lower()
        # raise an error similar to the one the default certificate verification would throw
        if self._expected_thumbprint != sha1Digest:
            raise vcops_pak_six.VcopsURLError(ssl.SSLError(
                '[SSL: CERTIFICATE_VERIFY_FAILED] expected certificate thumbprint not matched. Expected "{}", got "{}"'.format(
                    self._expected_thumbprint, sha1Digest)))

    def connect(self):
        """
        connect as usual (cert and hostname checks must be disabled), then check the cert by thumbprint
        """
        vcops_pak_six.VcopsHTTPSConnection.connect(self)
        self._validate_thumbprint()


class ThumbprintHTTPSHandler(vcops_pak_six.VcopsHTTPSHandler):
    """
    enable connection validation using a thumbprint
    subclass HTTPSHandler in order to pass ThumbprintHTTPSConnection to do_open

    Add thumbprint argument to __init__
    Change https_open to pass ThumbprintHTTPSConnection to do_open
    """
    _expected_thumbprint = None

    def __init__(self, debuglevel=0, context=None, thumbprint=None):
        if thumbprint:
            self._expected_thumbprint = thumbprint.replace(' ', '').replace(':', '').lower()
        vcops_pak_six.VcopsHTTPSHandler.__init__(self, debuglevel, context)

    def https_open(self, req):
        """
        Change the call to do_open.
        Create a context if necessary
        Preserve the settings on the context and set them back after making the connection.
        """
        # preserve any values set on the sslcontext if we have a thumbprint
        if self._expected_thumbprint:
            # pylint:disable=access-member-before-definition
            if not self._context:
                # pylint:enable=access-member-before-definition
                self._context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
            saved_verify_mode = self._context.verify_mode
            saved_hostname_mode = self._context.check_hostname
            self._context.check_hostname = False
            self._context.verify_mode = ssl.CERT_NONE

        # override with thumbprint checking connection
        try:
            rc = vcops_pak_six.VcopsHTTPSHandler.do_open(self, ThumbprintHTTPSConnection, req, context=self._context,
                                                     thumbprint=self._expected_thumbprint)

        finally:
            # and set the values back like we found them
            if self._expected_thumbprint and saved_verify_mode:
                self._context.verify_mode = saved_verify_mode
                self._context.check_hostname = saved_hostname_mode

        return rc


def wrapIPv6AddressInBrackets(url):
    """
    Imperfect check to detect IPv6 addresses will work for this hack
    @param url: URL that will have host wrapped in brackets if it is ipv6
    Returns the URL with host wrapped in brackets if ipv6
    """
    urlParts = url.split("/", 3)
    if len(urlParts) == 4:
        host = urlParts[2]
        if "[" not in host and "]" not in host and ":" in host and not host.endswith(":443") and not host.endswith(
                ":80"):
            scheme = urlParts[0]
            remainingUrl = urlParts[3]
            wrappedUrl = scheme + "//[" + host + "]/" + remainingUrl
            logging.debug("Wrapped ipv6 URL: " + wrappedUrl)
            return wrappedUrl
    return url


def create_cluster_ssl_context():
    """
    create an ssl context that validates server certificates based on the cluster truststore
    instead of the system default
    """
    # this gets called while upgrading from versions with python < 2.7.9, so the feature might not be present
    if not hasattr(ssl, 'SSLContext'):
        return None

    cafile = os.path.join(vcops_pack_properties.getStorage(), 'vcops', 'user', 'conf', 'ssl', 'web_chain.pem')
    with open(cafile, 'r') as f:
        cadata = f.read()
    if len(cadata) == 0:
        cafile = os.path.join(vcops_pack_properties.getStorage(), 'vcops', 'user', 'conf', 'ssl', 'web_cert.pem')
    sslcontext = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH, cafile=cafile)
    # not ideal - but the alternative is to check vs. thumbprints gathered from the same place and the trusted cert list
    sslcontext.check_hostname = False
    return sslcontext


def create_cluster_opener():
    """
    create an opener that validates server certificates based on the cluster truststore instead of the system default
    """
    sslcontext = create_cluster_ssl_context()
    # Force it to NOT use a proxy
    proxy_handler = vcops_pak_six.VcopsProxyHandler({})
    parameters = {}
    if sslcontext:
        parameters['context'] = sslcontext
    https_handler = vcops_pak_six.VcopsHTTPSHandler(**parameters)
    opener = vcops_pak_six.vcops_build_opener(proxy_handler, https_handler)
    return opener


def create_unverified_opener():
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    proxy_handler = vcops_pak_six.VcopsProxyHandler({})
    parameters = {}
    if ctx:
        parameters['context'] = ctx
    https_handler = vcops_pak_six.VcopsHTTPSHandler(**parameters)
    opener = vcops_pak_six.vcops_build_opener(proxy_handler, https_handler)
    return opener


def create_thumbprint_opener(thumbprint):
    """
    create an opener that verifies the server certificate by thumbprint
    """
    proxy_handler = vcops_pak_six.VcopsProxyHandler({})
    https_handler = ThumbprintHTTPSHandler(thumbprint=thumbprint)
    opener = vcops_pak_six.vcops_build_opener(proxy_handler, https_handler)
    return opener


def create_casa_url(url, hostname=None):
    if not hostname:
        hostname = 'localhost'

    return "https://{0}/casa{1}".format(hostname, url)


def create_bearer_auth_header(token):
    return "Bearer " + token


def make_auth_header(username, password):
    if vcops_pak_six.VCOPS_PY3:
        try:
            up = username + ":" + password
        except TypeError as e:
            username = str(username, 'utf-8')
            password = str(password, 'utf-8')
            up = username + ":" + password

        up = up.encode('utf-8')
        # pylint:disable=no-member
        authToken = base64.b64encode(up)
        # pylint:enable=no-member
        authToken = authToken.decode('utf-8')
        authHeader = 'Basic {0}'.format(authToken)
    else:
        up = '{0}:{1}'.format(username, password)
        authToken = base64.encodestring(up)[:-1] # strip appended newline
        authHeader = "Basic {0}".format(authToken)

    return authHeader


def create_opener(verify=True, thumbprint=None):
    if verify:
        if thumbprint:
            return create_thumbprint_opener(thumbprint)
        else:
            return create_cluster_opener()
    else:
        return create_unverified_opener()


def login(hostname,
          refresh_token=None,
          credential_retriever=None,
          verify=True,
          thumbprint=None):
    authorization_url = create_casa_url("/authorize", hostname)
    authorization_request = vcops_pak_six.VcopsRequest(authorization_url)
    authorization_request.get_method = lambda: 'POST'
    authorization_request.add_header('Content-Type', 'application/json')
    authorization_request.add_header('Accept', 'application/json')
    common_error_message = "Failed to get CaSA token pair"
    common_success_message = "Successfully retrieved casa token pair"

    try:
        opener = create_opener(verify, thumbprint)

        if refresh_token:
            try:
                auth_header = create_bearer_auth_header(refresh_token)
                authorization_request.add_header(Authorization, auth_header)
                response = opener.open(authorization_request)
                if response.getcode() == HTTP_OK:
                    logging.info(common_success_message)
                return response.getcode(), json.loads(response.read())
            except vcops_pak_six.VcopsHTTPError as e:
                if e.code == UNAUTHORIZED:
                    logging.warning(
                        "{0}: refresh token is expired or invalid for host {1}".format(common_error_message, hostname))
                elif e.code == NOT_FOUND:
                    logging.warning(
                        "{0}: token authentication is not supported by the host {1},".format(common_error_message, hostname))
                    return e.code, {}
                else:
                    logging.error("{0} for the host {1}, error code {2}".format(common_error_message, hostname, e.code))
                    return e.code, {}

        if credential_retriever:
            username, password = credential_retriever()
        else:
            username, password = vc_ops_utilities.returnMaintenanceUserCreds()

        auth_header = make_auth_header(username, password)
        authorization_request.add_header(Authorization, auth_header)

        response = opener.open(authorization_request)
        response_code = response.getcode()
        response_body = json.loads(response.read().decode('utf-8'))
        logging.info(common_success_message)

        return response_code, response_body

    except vcops_pak_six.VcopsHTTPError as e:
        if e.code != UNAUTHORIZED and e.code != NOT_FOUND:
            logging.error("{0}: '{1}' error code: {2}".format(common_error_message, e.code, e))
        return e.code, {}


def logout(hostname,
           token_pair,
           verify=True,
           thumbprint=None):
    if not token_pair or token_pair == {}:
        return HTTP_OK

    try:
        refresh_token = token_pair[REFRESH_TOKEN]
    except KeyError as e:
        return HTTP_OK

    if refresh_token == "" or refresh_token is None:
        return HTTP_OK

    invalidation_url = create_casa_url("/authorize", hostname)
    invalidation_request = vcops_pak_six.VcopsRequest(invalidation_url)
    invalidation_request.get_method = lambda: 'DELETE'
    try:
        opener = create_opener(verify, thumbprint)
        auth_header = create_bearer_auth_header(refresh_token)
        invalidation_request.add_header(Authorization, auth_header)
        response = opener.open(invalidation_request)
        return response.getcode()
    except vcops_pak_six.VcopsHTTPError as e:
        # in case if this token is already invalide
        if e.code == UNAUTHORIZED:
            return HTTP_OK
        logging.error("Failed to invalidate refresh token: status code: '{0}' exception: {1}".format(e.code, e))
        return e.code


def runHttpRequest(url,
                   request,
                   timeout=None,
                   verify=True,
                   authorization_required=True,
                   token_pair=None,
                   credential_retriever=None,
                   thumbprint=None,
                   suppress_data_logging=False):
    """
    Execute an Http request and returns: response code, result body
    @param url: URL for the request
    @param request: Request object
    @param timeout: Call timeout in seconds, None for no timeout.
    @param verify: is certificate verification needed
    @param authorization_required: is authorization header needed
    @param thumbprint:
    @param suppress_data_logging: if True, do not log the result body.
           Use this if the body contains sensitive information.
    @return A tuple containing http code or one of the HttpRequestErrors in case of exception and a result string
    :rtype tuple[int, str]
    """
    access_token_used = False
    response_body = ''
    auth_header_set_externally = None
    token_pair_set_externally = None
    try:
        opener = create_opener(verify, thumbprint)

        auth_header_set_externally = request.has_header(Authorization)
        token_pair_set_externally = True if token_pair else False
        # remove "Authorization" header for unauthorized APIs
        if auth_header_set_externally:
            if not authorization_required:
                request.remove_header(Authorization)
            elif request.get_header(Authorization).startswith("Bearer"):
                access_token_used = True

        if not auth_header_set_externally and authorization_required:
            if not token_pair_set_externally:
                response_status, token_pair = login(request.host,
                                                    verify=verify,
                                                    credential_retriever=credential_retriever,
                                                    thumbprint=thumbprint)

                if response_status == NOT_FOUND:
                    if credential_retriever:
                        username, password = credential_retriever()
                    else:
                        username, password = vc_ops_utilities.returnMaintenanceUserCreds()
                    auth_header = make_auth_header(username, password)
                    request.add_header(Authorization, auth_header)
                elif response_status == HTTP_OK:
                    access_token_used = True
                    auth_header = create_bearer_auth_header(token_pair[ACCESS_TOKEN])
                    request.add_header(Authorization, auth_header)
            else:
                access_token_used = True
                auth_header = create_bearer_auth_header(token_pair[ACCESS_TOKEN])
                request.add_header(Authorization, auth_header)

        if timeout:
            r = opener.open(request, None, timeout)
        else:
            r = opener.open(request)

        result = r.read()
        response_code = r.getcode()
        if vcops_pak_six.VCOPS_PY3:
            response_body = result.decode('utf-8')
        else:
            response_body = result

        if suppress_data_logging:
            logging.debug('Code: {0} - Body: <masked>'.format(r.getcode()))
        else:
            logging.debug('Code: {0} - Body: {1}'.format(r.getcode(), result))
    except vcops_pak_six.VcopsHTTPError as e:
        logging.warning("Open URL: '{0}' returned code: {1}, exception: {2!s}".format(url, e.code, e))
        if e.code == UNAUTHORIZED and access_token_used and token_pair:
            response_status, token_pair = login(request.host,
                                                verify=verify,
                                                refresh_token=token_pair[REFRESH_TOKEN],
                                                credential_retriever=credential_retriever,
                                                thumbprint=thumbprint)
            if response_status in (HTTP_OK, NOT_FOUND):
                if response_status == HTTP_OK:
                    auth_header = create_bearer_auth_header(token_pair[ACCESS_TOKEN])
                    request.add_header(Authorization, auth_header)
                elif response_status == NOT_FOUND:
                    request.remove_header(Authorization)

                return runHttpRequest(url=url,
                                      request=request,
                                      timeout=timeout,
                                      verify=verify,
                                      authorization_required=authorization_required,
                                      token_pair=token_pair,
                                      credential_retriever=credential_retriever,
                                      thumbprint=thumbprint,
                                      suppress_data_logging=suppress_data_logging)
        response_code = e.code
    except vcops_pak_six.VcopsURLError as e:
        logging.warning("Open URL: '{0}' failed with a URLERROR returned reason: {1}, exception: {2!s}".format(
            url, e.reason, e))
        response_code = HttpRequestErrors.URLLIB2_URLERROR
    except socket.timeout as e:
        logging.warning('Open URL: "{0}" failed with a socket.timeout, exception: {1!s}'.format(url, e))
        response_code = HttpRequestErrors.SOCKET_TIMEOUT
    except ssl.SSLError as e:
        logging.warning('Open URL: "{0}" failed with an ssl.SSLError, exception: {1!s}'.format(url, e))
        response_code = HttpRequestErrors.SSL_SSLERROR
    except vcops_pak_six.VcopsHTTPException as e:
        logging.warning('Open URL: "{0}" failed with an HTTPException, exception: {1!s}'.format(url, e))
        response_code = HttpRequestErrors.HTTPLIB_HTTPEXCEPTION

    # in case if token auth header set internally, we need to invalidate it after usage
    if not auth_header_set_externally and not token_pair_set_externally:
        logout(request.host, token_pair, verify=verify, thumbprint=thumbprint)

    logging.debug("Open URL: '{0}' response code: {1}".format(url, response_code))

    return response_code, response_body, token_pair


def getUrl(url,
           timeout=None,
           additional_headers=None,
           verify=True,
           authorization_required=True,
           token_pair=None,
           credential_retriever=None,
           thumbprint=None,
           suppress_data_logging=False):
    """
    GET a URL and returns: response code, result body
    @param url: URL for the request
    @param timeout: Call timeout in seconds, None for no timeout.
    @param additional_headers: Additional request headers
    @param suppress_data_logging: if True, do not log the result body.
           Use this if the body contains sensitive information.
    @return A tuple containing http code or one of the HttpRequestErrors in case of exception and a result string
    :param verify:
    :rtype tuple[int, str]
    """

    if additional_headers is None:
        additional_headers = {}
    url = wrapIPv6AddressInBrackets(url)
    req = vcops_pak_six.VcopsRequest(url, None, additional_headers)
    return runHttpRequest(url,
                          req,
                          timeout,
                          verify=verify,
                          authorization_required=authorization_required,
                          token_pair=token_pair,
                          credential_retriever=credential_retriever,
                          thumbprint=thumbprint,
                          suppress_data_logging=suppress_data_logging)


def postUrl(url,
            values,
            timeout=None,
            additional_headers=None,
            verify=True,
            authorization_required=True,
            token_pair=None,
            credential_retriever=None,
            thumbprint=None,
            suppress_data_logging=False):
    """
    POST to a URL and returns: responseCode, result body
    @param url: URL for the request
    @param values: The values to be sent in the request. The method will encode them as JSON.
    @param timeout: Call timeout in seconds, None for no timeout.
    @param additional_headers: Additional request headers
    @param suppress_data_logging: if True, do not log the result body.
           Use this if the body contains sensitive information.
    @return A tuple containing http code or one of the HttpRequestErrors in case of exception and a result string
    :param verify:
    :param opener:
    :rtype tuple[int, str]
    """

    if additional_headers is None:
        additional_headers = {}

    if vcops_pak_six.VCOPS_PY3:
        jdata = json.dumps(values).encode('utf-8')
    else:
        jdata = json.dumps(values)

    url = wrapIPv6AddressInBrackets(url)
    req = vcops_pak_six.VcopsRequest(url, jdata, additional_headers)
    req.add_header("Content-Type", "application/json")

    return runHttpRequest(url,
                          req,
                          timeout=timeout,
                          verify=verify,
                          authorization_required=authorization_required,
                          token_pair=token_pair,
                          credential_retriever=credential_retriever,
                          thumbprint=thumbprint,
                          suppress_data_logging=suppress_data_logging)


def postMultipartUrl(url,
                     values,
                     timeout=None,
                     additional_headers=None,
                     verify=True,
                     authorization_required=True,
                     token_pair=None,
                     credential_retriever=None,
                     thumbprint=None,
                     suppress_data_logging=False):
    """
    POST to a URL and returns: responseCode, result body
    @param url: URL for the request
    @param values: The values to be sent in the request. The method will encode them as JSON.
    @param timeout: Call timeout in seconds, None for no timeout.
    @param additional_headers: Additional request headers
    @param suppress_data_logging: if True, do not log the result body.
           Use this if the body contains sensitive information.
    @return A tuple containing http code or one of the HttpRequestErrors in case of exception and a result string
    :param verify:
    :rtype tuple[int, str]
    """

    if additional_headers is None:
        additional_headers = {}

    url = wrapIPv6AddressInBrackets(url)
    req = vcops_pak_six.VcopsRequest(url, values, additional_headers)

    return runHttpRequest(url,
                          req,
                          timeout=timeout,
                          verify=verify,
                          authorization_required=authorization_required,
                          token_pair=token_pair,
                          credential_retriever=credential_retriever,
                          thumbprint=thumbprint,
                          suppress_data_logging=suppress_data_logging)


def putUrl(url,
           values,
           timeout=None,
           additional_headers=None,
           verify=True,
           authorization_required=True,
           token_pair=None,
           credential_retriever=None,
           thumbprint=None,
           suppress_data_logging=False):
    if additional_headers is None:
        additional_headers = {}

    if vcops_pak_six.VCOPS_PY3:
        jdata = json.dumps(values).encode('utf-8')
    else:
        jdata = json.dumps(values)

    url = wrapIPv6AddressInBrackets(url)
    req = vcops_pak_six.VcopsRequest(url, jdata, additional_headers)
    req.add_header("Content-Type", "application/json")
    req.get_method = lambda: 'PUT'
    return runHttpRequest(url,
                          req,
                          timeout=timeout,
                          verify=verify,
                          authorization_required=authorization_required,
                          token_pair=token_pair,
                          credential_retriever=credential_retriever,
                          thumbprint=thumbprint,
                          suppress_data_logging=suppress_data_logging)
