#!/usr/bin/python -cprint 'ERROR: You must invoke this script using $VMWARE_PYTHON_BIN'; import sys; sys.exit(5)
# Copyright 2014 VMware, Inc.  All rights reserved. -- VMware Confidential
#
# Simplified access to various vCOps Suite properties, in particular to often-used ones.


__author__ = "VMware, Inc."

import logging
import os
import re

try:
    vcops_base_dir = os.environ["ALIVE_BASE"]
except KeyError:
    vcops_base_dir = os.path.join(os.sep, "usr", "lib", "vmware-vcops")

utilitiesDir = os.path.join(vcops_base_dir, "..", "vmware-vcopssuite", "utilities")

import vcops_pak_six

class PropertyError(Exception):
   """
   Used to describe errors that occurred while retrieving properties
   """
   def __init__(self, value, code):
      self.value = value
      self.code = code

   def __str__(self):
      return self.value


def _load_property_value(file, key, raiseError=False, logging_enabled=True):
    """
    Return a value for given key read from a properties file.
    @param file - file to load
    @param key - key to search
    @raiseError - Will raise PropertyError with an appropriate error code if set to true when key is not found
                  or file not found or file could not be read
    @param logging_enabled - Whether or not this function should perform any logging. Defaults to True.
    @return  The value associated to the key, or None if not found and raiseError is false.
    """
    keyString = key + '\s*='
    reSearchPattern = r'^' + keyString + '.*$'
    if not os.path.exists(file):
        if logging_enabled:
            logging.error('File: "{0}" does not exist'.format(file))
        if raiseError:
            raise PropertyError('Properties file does not exist', 1)
    keyReturnedString = None
    try:
        with open(file, 'r') as workingFile:
            for line in workingFile:
                keyReturnedString = re.search(reSearchPattern, line)
                if keyReturnedString:
                    break
    except EnvironmentError as e:
        if logging_enabled:
            logging.error('Failed to open the file: "{0}" caused by exception: {1}'.format(file, str(e)))
        if raiseError:
            raise PropertyError('Failed to open properties file', 1)
    if keyReturnedString:
        valueString = keyReturnedString.group(0).split('=', 1)[1].strip()
        return valueString
    else:
        if logging_enabled:
            logging.error('Failed to find the key: "{0}" in the file: "{1}"'.format(key, file))
        if raiseError:
            raise PropertyError('Failed to find key in property file', 1)
        else:
            return None


def _load_ini_value(filePath, section, key, raiseError=False):
    """
    Returns a dictionary value from an INI-style config file
    @param file - file to load
    @param key - key to search
    @raiseError - Will raise PropertyError with an appropriate error code if set to true when key is not found
                  or file not found or file could not be read
    @return  The value associated to the key, or None if not found and raiseError is false.
    """
    logging.debug("Return value: section: '{0}', key '{1}', file '{2}'".format(section, key, filePath))
    config = vcops_pak_six.configparser.RawConfigParser()
    try:
        config.read(filePath)
    except EnvironmentError as e:
        logging.warn("Failed to open '{0}': {1}".format(filePath, e))
        # Already logged by openFile
        if raiseError:
           raise PropertyError('Properties file does not exist', 1)
        return ''

    returnValue = None
    try:
        returnValue = config.get(section, key)
    except vcops_pak_six.configparser.NoOptionError as e:
        logging.warning("Failed to return value: file '{0}', section '{1}', key '{2}': {3}"
                        .format(filePath, section, key, e))
        if raiseError:
            raise PropertyError('Option does not exist', 1)
        else:
            return None
    except vcops_pak_six.configparser.NoSectionError as e:
        logging.warning("Failed to return value: file '{0}', section '{1}', key '{2}': {3}"
                        .format(filePath, section, key, e))
        if raiseError:
           raise PropertyError('Section does not exist', 1)
        else:
            return None

    return returnValue


def getPlatformStatePropertyFile():
    """Return the filename of the vCenter Operations Suite properties file."""

    return os.path.join(utilitiesDir, 'sliceConfiguration', 'data', 'platformState.properties')


def getSuitePropertyFile():
    """Return the filename of the vCenter Operations Suite properties file."""

    return os.path.join(utilitiesDir, 'conf', 'vCenterOperationsSuite.properties')

def getSuiteProperty(propertyName, envVarName=None, defaultValue=None, logging_enabled=True):
    """Read the property <propertyName> from the vCOps Suite properties file. If it doesn't exist, try
       the <envVarName> variable from the environment (it defaults to <propertyName>). If that doesn't exist either, return the default value.
    """

    value = _load_property_value(getSuitePropertyFile(), propertyName, logging_enabled=logging_enabled)
    if not value:
        try:
            value = os.environ[envVarName or propertyName]
        except KeyError:
            if logging_enabled:
                logging.error("{0} not set in environment.".format(envVarName))
    if not value:
        value = defaultValue

    if logging_enabled:
        logging.debug("{0}: '{1}'".format(propertyName, value))
    return value


def getVcopsUser():
    """Read the VCOPS_USER from the environment."""

    return getSuiteProperty("ADMIN_USERNAME", "VCOPS_USER", "admin")


def getVcopsGroup():
    """Read the VCOPS_GROUP from the environment."""

    return getSuiteProperty("VCOPS_GROUP", "VCOPS_GROUP", getVcopsUser())


def getVcopsUserGroup():
    """Read the VCOPS_USER and VCOPS_GROUP from the environment, and return it as (user, group) tuple."""

    vcopsUser = getSuiteProperty("ADMIN_USERNAME", "VCOPS_USER")
    vcopsGroup = getSuiteProperty("VCOPS_GROUP", "VCOPS_GROUP", vcopsUser)

    return (vcopsUser, vcopsGroup)


def getStorage():
    """Read the STORAGE from the environment."""

    return getSuiteProperty("STORAGE", "STORAGE", os.path.join(os.path.sep, "data"))


def getJavaHome():
    """Read the JAVA_HOME from the environment."""

    return getSuiteProperty("JAVA_HOME")


def getVmwareJavaHome():
    """Read the VMWARE_JAVA_HOME from the environment."""

    return getSuiteProperty("VMWARE_JAVA_HOME")


def getCatalinaHome():
    """Read the CATALINA_HOME from the environment."""

    return getSuiteProperty("CATALINA_HOME")

def getPhotonApacheHome():
    """Read the PHOTON_APACHE_HOME from the environment."""

    return getSuiteProperty("PHOTON_APACHE_HOME")

def getApacheHome():
    """Read the APACHE_HOME from the environment."""

    return getSuiteProperty("APACHE_HOME")

def getVcopsBase():
    """Read the VCOPS_BASE from the environment."""

    return getSuiteProperty("VCOPS_BASE")

def getPythonBin():
    """Get the VMWARE_PYTHON_BIN property from the utilities property file"""

    return getSuiteProperty("VMWARE_PYTHON_BIN")

def getPython3Bin():
    """Get the VMWARE_PYTHON_3_BIN property from the utilities property file"""

    return getSuiteProperty("VMWARE_PYTHON_3_BIN")

def getSliceUuid():
    """Return the slice instance uuid."""

    return _load_ini_value(getPlatformStatePropertyFile(), "ids", "slicedinstanceid")

def getSslDir():
    """
    compute the target folder for certificates and links
    @Return full path to ssl directory
    """
    return os.path.join(getStorage(), 'vcops', 'user', 'conf', 'ssl')
